//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#include "StdAfx.h"
#include "zlib/zlib.h"
#include "updownclient.h"
#include "PartFile.h"
#include "opcodes.h"
#include "packets.h"
#include "emule.h"
#include "otherfunctions.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

//	members of CUpDownClient
//	which are mainly used for downloading functions

void CUpDownClient::DrawStatusBar(CDC* dc, RECT* rect, bool bOnlyGreyRect, bool bFlat)
{
	EMULE_TRY

	const COLORREF crPending = RGB(222, 160, 0);
	const COLORREF crToDo = RGB(255, 208, 0);
	const COLORREF crNextPending = RGB(255, 255, 100);
	const COLORREF crMeOnly = RGB(190, 190, 255);
	COLORREF crBoth, crNeither, crClientOnly;

	if (bFlat)
	{
		crBoth = RGB(0, 150, 0);
		crNeither = RGB(224, 224, 224);
		crClientOnly = RGB(0, 0, 0);
	}
	else
	{
		crBoth = RGB(0, 192, 0);
		crNeither = RGB(240, 240, 240);
		crClientOnly = RGB(95, 95, 95);
	}

	uint32		filesize = (m_pReqPartFile) ? m_pReqPartFile->GetFileSize() : 1;
	CBarShader	statusBar(rect->bottom - rect->top, rect->right - rect->left, crNeither, filesize);

	if (!bOnlyGreyRect && (m_pReqPartFile != NULL) && m_pbytePartStatuses)
	{
		_Bvector gettingParts;
		bool isDLing = (m_eDownloadState == DS_DOWNLOADING);

		if (isDLing)
			ShowDownloadingParts(gettingParts);

		uint32 dwStart, dwEnd;

		for (uint32 i = 0; i < m_uPartCount; i++)
		{
			bool bIsComplete = m_pReqPartFile->IsPartComplete(i, &dwStart, &dwEnd);

			dwEnd++;

		//	They have this part. Do I have it?
			if (m_pbytePartStatuses[i])
			{ 
				if (bIsComplete)
					statusBar.FillRange(dwStart, dwEnd, crBoth);
				else if (isDLing && (m_dwLastBlockOffset < dwEnd) && (m_dwLastBlockOffset >= dwStart))
				{
					statusBar.FillRange(dwStart, m_dwLastBlockOffset, crPending);
					statusBar.FillRange(m_dwLastBlockOffset, dwEnd, crToDo);
				}
				else if (isDLing && gettingParts[i])
					statusBar.FillRange(dwStart, dwEnd, crNextPending);
				else
					statusBar.FillRange(dwStart, dwEnd, crClientOnly);
			}
		//	They don't have a part. Do I have it?
			else if (bIsComplete)
				statusBar.FillRange(dwStart, dwEnd, crMeOnly);
		}
	}

	statusBar.Draw(dc, rect->left, rect->top, bFlat);

	EMULE_CATCH
}

bool CUpDownClient::Compare(CUpDownClient* tocomp)
{
	EMULE_TRY

	if (GetIP() && tocomp->GetIP() && GetIP() == tocomp->GetIP() && GetUserPort() == tocomp->GetUserPort())
		return true;

	if (HasValidHash() && tocomp->HasValidHash())
		return !md4cmp(GetUserHash(), tocomp->GetUserHash());

	if (HasLowID())
		return((GetUserID() == tocomp->GetUserID()) && (GetServerIP() == tocomp->GetServerIP()) && (GetUserPort() == tocomp->GetUserPort()));
	else
		return((GetUserID() == tocomp->GetUserID() && GetUserPort() == tocomp->GetUserPort()) || (GetIP() && (GetIP() == tocomp->GetIP() && GetUserPort() == tocomp->GetUserPort())));

	EMULE_CATCH

	return false;
}

// Return TRUE if client still connected, otherwise return FALSE
bool CUpDownClient::AskForDownload()
{
	EMULE_TRY

#ifdef OLD_SOCKETS_ENABLED

	if (g_eMuleApp.m_pListenSocket->TooManySockets() && !(m_pRequestSocket && m_pRequestSocket->IsConnected()))
	{
		if (GetDownloadState() != DS_TOOMANYCONNS && GetDownloadState() != DS_DOWNLOADING)
			SetDownloadState(DS_TOOMANYCONNS);

	//	Asking time random only at the beginning for each new source
		if (m_dwLastAskedTime == 0)
		{
			m_dwLastAskedTime = GetTickCount() - FILEREASKTIME + static_cast<uint32>(ROUND((static_cast<double>(rand()) / RAND_MAX) * 300000.0));
		}

		return true;
	}

#endif //OLD_SOCKETS_ENABLED

	m_bUDPPending = false;

	m_dwLastAskedTime = GetTickCount();

	SetDownloadState(DS_CONNECTING);

	if (TryToConnect())
		return true;
	else
		return false;

	EMULE_CATCH

	return false;
}

bool CUpDownClient::IsSourceRequestAllowed()
{
	EMULE_TRY

	DWORD			dwTickCount = GetTickCount() + CONNECTION_LATENCY;
	unsigned int	nTimePassedClient = dwTickCount - GetLastSrcAnswerTime();
	unsigned int	nTimePassedFile = dwTickCount - m_pReqPartFile->GetLastAnsweredTime();
	bool			bNeverAskedBefore = GetLastSrcReqTime() == 0;

	UINT			dwSources = m_pReqPartFile->GetSourceCount();

//	Original client return false if client's SourceExchange Version is less or equal than 1
//	but i think we need to keep compatibility with old ePlus versions
	return (
			//	If client has the correct extended protocol
				ExtProtocolAvailable() && GetSourceExchangeVersion() >= 1 &&
			//	AND if we can send sources
				!g_eMuleApp.m_pGlobPrefs->IsDisabledXS() &&
			//	AND if we need more sources
				g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFileSoft() > m_pReqPartFile->GetSourceCount() &&
			//	AND if...
				(
				//	source is not complete and file is rare, allow once every 10 minutes
					( !m_bCompleteSource
						&& ( dwSources - m_pReqPartFile->GetValidSourcesCount() <= RARE_FILE / 4
							|| dwSources <= RARE_FILE * 2 )
						&& (bNeverAskedBefore || nTimePassedClient > SOURCECLIENTREASK) )
				//	otherwise, allow every 90 minutes, but only if we haven't
				//	asked someone else in last 10 minutes
					|| ( (bNeverAskedBefore || nTimePassedClient > (unsigned)(SOURCECLIENTREASK * m_pReqPartFile->GetCommonFilePenalty())) &&
						(nTimePassedFile > SOURCECLIENTREASK) )
				)
			);

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendFileRequest()
{
	EMULE_TRY

	ASSERT(m_pReqPartFile != NULL);

	if (m_pReqPartFile == NULL)
		return;

	AddAskedCountDown();

	CSafeMemFile		packetStream1(16 + 16);

	packetStream1.Write(m_pReqPartFile->GetFileHash(), 16);

	if (GetExtendedRequestsVersion() > 0)
		m_pReqPartFile->WritePartStatus(&packetStream1);

	if (GetExtendedRequestsVersion() > 1)
		m_pReqPartFile->WriteCompleteSourcesCount(&packetStream1); //netwolf: complete sources (zegzav) 13.05.03

	Packet		*pPacket = new Packet(&packetStream1);

	pPacket->m_eOpcode = OP_REQUESTFILENAME;

	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);

#ifdef OLD_SOCKETS_ENABLED
	m_pRequestSocket->SendPacket(pPacket, true);
#endif //OLD_SOCKETS_ENABLED

//	Removed requesting the file status for files <= PARTSIZE for better compatibility with ed2k protocol (eDonkeyHybrid).
//	if the remote client answers OP_REQUESTFILENAME with OP_REQFILENAMEANSWER the file is shared by the remote client. if we
//	know that the file is shared, we know also that the file is complete and don't need to request the file status.
	if (m_pReqPartFile->GetPartCount() > 1)
	{
		Packet 	*pPacket = new Packet(OP_SETREQFILEID, 16);

		md4cpy(pPacket->m_pcBuffer, m_pReqPartFile->GetFileHash());
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);

#ifdef OLD_SOCKETS_ENABLED

		m_pRequestSocket->SendPacket(pPacket, true);

#endif //OLD_SOCKETS_ENABLED
	}

	if (IsEmuleClient())
	{
		SetRemoteQueueFull(true);
		SetRemoteQueueRank(0);
	}

	if ( g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFileSoft() < m_pReqPartFile->GetSourceCount()
	  || g_eMuleApp.m_pGlobPrefs->IsDisabledXS() )
	{
		return;
	}
//	Disable XS for up to X sources
	if ( g_eMuleApp.m_pGlobPrefs->DisableXSUpTo()
	  && g_eMuleApp.m_pGlobPrefs->XSUpTo() < m_pReqPartFile->GetSourceCount() )
	{
		return;
	}

	if (IsSourceRequestAllowed())
	{
		m_pReqPartFile->SetLastAnsweredTimeTimeout();

		Packet		*pPacket = new Packet(OP_REQUESTSOURCES, 16, OP_EMULEPROT);

		md4cpy(pPacket->m_pcBuffer, m_pReqPartFile->GetFileHash());
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadSourceExchange(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
		m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	ProcessFileInfo() processes the contents of a File Request Answer packet
void CUpDownClient::ProcessFileInfo(char *pcPacketBuf, uint32 dwPacketSize)
{
	EMULE_TRY

//	Create a memory stream from the packet and read its contents
	CSafeMemFile		packetStream(reinterpret_cast<BYTE*>(pcPacketBuf), dwPacketSize);

	uchar		fileHash[16];

	packetStream.Read(fileHash, sizeof(fileHash));

	m_dwLastAskedTime = ::GetTickCount();

//	Patch 'fix disappearing m_pReqPartFile'
	if (m_pReqPartFile == NULL)
	{
	//	try to find the file with given hash in our DL list
		m_pReqPartFile = g_eMuleApp.m_pDownloadQueue->GetFileByID(fileHash);

	//	if the search was not succesfull, then we don't need this file anymore (perhaps it was canceled)
		if (m_pReqPartFile == NULL)
		{
			m_strClientFilename.Empty();
			AddDebugLogLine(RGB_LOG_DIMMED + _T("Client '%s' (%s): offered file not found (hash: %s)"), GetUserName(), GetClientNameAndVersionString(), HashToString(fileHash));
			return;
		}
	}
	
//	report about difference in requested & accepted hash
	if (md4cmp(fileHash, m_pReqPartFile->GetFileHash()))
	{
		CPartFile		*pReqFileDownloading = g_eMuleApp.m_pDownloadQueue->GetFileByID(fileHash);

		if (pReqFileDownloading)
			AddDebugLogLine(RGB_LOG_DIMMED +  _T("Client '%s' (%s): offered file was not requested (offered: %s | requested: %s)"), GetUserName(), GetClientNameAndVersionString(), pReqFileDownloading->GetFileName(), m_pReqPartFile->GetFileName());
		else
			AddDebugLogLine(RGB_LOG_DIMMED +  _T("Client '%s' (%s): offered file was not requested (offered hash: %s | requested file: %s)"), GetUserName(), GetClientNameAndVersionString(), HashToString(fileHash), m_pReqPartFile->GetFileName());
	}
//	otherwise process the rest of the packet 
	else	
	{
		uint16		uNameLength;

		packetStream.Read(&uNameLength, sizeof(uint16));

		packetStream.Read(m_strClientFilename.GetBuffer(uNameLength), uNameLength);
		m_strClientFilename.ReleaseBuffer(uNameLength);

	//	Removed requesting the file status for files <= PARTSIZE for better compatibility with ed2k protocol (eDonkeyHybrid).
	//	if the remote client answers OP_REQUESTFILENAME with OP_REQFILENAMEANSWER the file is shared by the remote client. if we
	//	know that the file is shared, we know also that the file is complete and don't need to request the file status.

		if (m_pReqPartFile->GetPartCount() == 1)
		{
			if (m_pbytePartStatuses)
			{
				delete[] m_pbytePartStatuses;
				m_pbytePartStatuses = NULL;
			}

			m_uPartCount = m_pReqPartFile->GetPartCount();
			m_pbytePartStatuses = new byte[m_uPartCount];
			memset2(m_pbytePartStatuses, 1, m_uPartCount);
			m_bCompleteSource = true;

		//	even if the file is <= PARTSIZE, we _may_ need the hashset for that file (if the file size == PARTSIZE)
			if (m_pReqPartFile->m_bHashSetNeeded)
				SendHashsetRequest();
			else 
				SendStartupLoadRequest();

			m_pReqPartFile->NewSrcPartsInfo();
			UpdateDisplayedInfo();
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::ProcessFileStatus(CSafeMemFile *packetStream, CPartFile *pReqPartFile, bool bUDPPacket/* = false*/)
{
	EMULE_TRY

	if (m_pReqPartFile == NULL)
	{
		m_pReqPartFile = pReqPartFile;

		if (m_pReqPartFile == NULL)
		{
//		This line not needed as mostly paired with other error stating just something similar
//			AddDebugLogLine( RGB_LOG_DIMMED + _T("Client '%s' (%s): no file stated"), GetUserName(),
//							 GetClientNameAndVersionString() );
			return;
		}
	}
	else if (m_pReqPartFile != pReqPartFile)
	{
//		This line not needed as mostly paired with other error stating just something similar
//		AddDebugLogLine( RGB_LOG_DIMMED + _T("Client '%s' (%s): wrong file stated (%s)"), GetUserName(),
//						 GetClientNameAndVersionString(), pReqPartFile ? pReqPartFile->GetFileName() : _T("NULL") );
		return;
	}

	uint16 uED2KPartCount;

	packetStream->Read(&uED2KPartCount, 2);

	if (m_pbytePartStatuses)
	{
		delete[] m_pbytePartStatuses;
		m_pbytePartStatuses = NULL;
	}

	bool bPartsNeeded = false;

	m_uPartCount = m_pReqPartFile->GetPartCount();
	if (uED2KPartCount == 0)
	{
		m_pbytePartStatuses = new byte[m_uPartCount];
		memset2(m_pbytePartStatuses, 1, m_uPartCount);
		bPartsNeeded = true;
		m_bCompleteSource = true;
	}
	else
	{
		if (m_pReqPartFile->GetED2KPartCount() != uED2KPartCount)
		{
			m_uPartCount = 0;
			CString strTemp;

			strTemp.Format(_T("wrong part number, file: %s (ED2KParts/Parts: %u/%u"), m_pReqPartFile->GetFileName(), m_pReqPartFile->GetED2KPartCount(), uED2KPartCount);
			throw strTemp;
		}

		m_bCompleteSource = false;
		uint16 nCompleteSourcePart = 0;
		m_pbytePartStatuses = new byte[m_uPartCount];
		uint16 done = 0;

		while (done != m_uPartCount)
		{
			byte toread;
			packetStream->Read(&toread, 1);

			for (sint32 i = 0;i != 8;i++)
			{
				m_pbytePartStatuses[done] = ((toread >> i) & 1) ? 1 : 0;
				if (m_pbytePartStatuses[done])
				{
					nCompleteSourcePart++;
					if (!m_pReqPartFile->IsPartComplete(done))
						bPartsNeeded = true;
				}
				done++;

				if (done == m_uPartCount)
					break;
			}
		}

		if (nCompleteSourcePart == m_uPartCount)
			m_bCompleteSource = true;
	}

//	if it is TCP connection, we can send something back
	if (!bUDPPacket)
	{
		if (m_pReqPartFile->m_bHashSetNeeded)
			SendHashsetRequest();
	//	request the file & enter to the queue only if we required a file from remote client
		else if (bPartsNeeded)
			SendStartupLoadRequest();
		else
			SetDownloadState(DS_NONEEDEDPARTS);
	}
//	if it was the UDP client than we have them on the queue
	else
	{
		if (!bPartsNeeded)
			SetDownloadState(DS_NONEEDEDPARTS);
		else if (bPartsNeeded && GetDownloadState() != DS_ONQUEUE)
			SetDownloadState(DS_ONQUEUE);
	}

	m_pReqPartFile->NewSrcPartsInfo();
	UpdateDisplayedInfo();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUpDownClient::AddRequestForAnotherFile(CPartFile* file)
{
	EMULE_TRY

	if (file == NULL)
		return false;

//	Check Other list
	if (m_otherRequestsList.Find(file) != 0)
		return false;
	else
	m_otherRequestsList.AddTail(file);

	if (!file->m_A4AFsrcList.Find(this))
		file->m_A4AFsrcList.AddTail(this);

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::RemoveRequestForAnotherFile(CPartFile* pPartFile)
{
	EMULE_TRY

	if (pPartFile == NULL)
		return;

//	Check Other list
	POSITION posPartFile = m_otherRequestsList.Find(pPartFile);

	if (posPartFile != NULL)
	{
		m_otherRequestsList.RemoveAt(posPartFile);
	}

	m_otherNoNeededMap.RemoveKey(pPartFile);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetDownloadState(EnumDLQState byNewState)
{
	EMULE_TRY

	if (m_eDownloadState != byNewState)
	{
	//	Clear variables before switching
		switch (m_eDownloadState)
		{
			case DS_DOWNLOADING:
				if (m_pReqPartFile)
					m_pReqPartFile->RemoveDownloadingSource(this);

				if (m_bTransferredDownMini && byNewState != DS_ERROR)
					g_eMuleApp.m_pGlobPrefs->Add2DownSuccessfulSessions();
				else
					g_eMuleApp.m_pGlobPrefs->Add2DownFailedSessions();

				g_eMuleApp.m_pGlobPrefs->Add2DownSAvgTime((::GetTickCount() - m_dwDownStartTime) / 1000);

			//	Prevent any action from the client
				m_eDownloadState = byNewState;

				ClearDownloadBlocksList();
				ClearPendingBlocksList();

				m_dwDownDataRate = 0;
				m_dwClientSumDLDataRateOverLastNMeasurements = 0;
				m_dwClientSumDLTickOverLastNMeasurements = 0;
				m_averageDLTickList.clear();
				m_averageDLDataRateList.clear();

				if (byNewState == DS_NONE)
				{
					if (m_pbytePartStatuses)
						delete[] m_pbytePartStatuses;

					m_pbytePartStatuses = 0;

					m_uPartCount = 0;
				}

				m_uLastPartAsked = 0xFFFF;

#ifdef OLD_SOCKETS_ENABLED

				if (m_pRequestSocket && byNewState != DS_ERROR)
					m_pRequestSocket->DisableDownloadLimit();
#endif //OLD_SOCKETS_ENABLED
				break;

			case DS_ERROR:
				return;
		}

	//	Switch the state & update info
		m_eDownloadState = byNewState;

	//	Prepare variables for new state
		switch (byNewState)
		{
			case DS_DOWNLOADING:
			//	Update statistical variables (start time of DL & minimal amount)
				m_dwLastDataRateCalculationTime = m_dwDownStartTime = ::GetTickCount();
				m_dwLastBlockOffset = ~0;
				m_bTransferredDownMini = false;

				m_pReqPartFile->AddDownloadingSource(this);

				if (IsEmuleClient())
					SetRemoteQueueFull(false);

				SetAskedCountDown(0);

				SetRemoteQueueRank(0, false);

				break;
			case DS_CONNECTED:
				m_dwEnteredConnectedState = ::GetTickCount();
				break;
		}

		UpdateDisplayedInfo();
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::ProcessHashSet(char* pPacket, uint32 size)
{
	EMULE_TRY

	if (!m_fRequestingHashSet)
		throw CString(_T("unwanted hashset received"));

	if ((!m_pReqPartFile) || md4cmp(pPacket, m_pReqPartFile->GetFileHash()))
		throw CString(_T("wrong fileID sent, file: ") + ((m_pReqPartFile != NULL) ? m_pReqPartFile->GetFileName() : _T("NULL")));

	CSafeMemFile data1((BYTE*)pPacket, size);

	if (m_pReqPartFile->LoadHashsetFromFile(data1, true))
	{
		m_fRequestingHashSet = 0;

		SendStartupLoadRequest();
	}
	else
	{
		m_pReqPartFile->m_bHashSetNeeded = true;
		AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_BADHASHSET), m_pReqPartFile->GetFileName());
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendBlockRequests()
{
	EMULE_TRY

	m_dwLastBlockReceived = ::GetTickCount();

	if (m_pReqPartFile == NULL)
		return;

//	Get blocks from the same chunk first
	if (m_downloadBlocksList.IsEmpty())
	{
	//	Get part for request. If actual part still has empty blocks, the we get that part,
	//	otherwise it will be a new part.
		m_uLastPartAsked = m_pReqPartFile->GetNextRequiredPart(this);

	//	Previous active part unknown
		if (m_uLastPartAsked != 0xFFFF)
		{
			uint32	dwCount, dwSavedCount;

			dwSavedCount = 3 - m_pendingBlocksList.GetCount();

			Requested_Block_Struct	**ppRequestedBlocksToAdd = new Requested_Block_Struct * [dwSavedCount];

			dwCount = m_pReqPartFile->GetGapsInPart(m_uLastPartAsked, ppRequestedBlocksToAdd, dwSavedCount);

		//	Finished block (part ?)
			if (dwCount == 0)
			{
			//	Check if part was completed
				if ( m_pReqPartFile->IsPartFull(m_uLastPartAsked)
				  || m_pReqPartFile->IsCorruptedPart(m_uLastPartAsked) )
				{
					m_uLastPartAsked = 0xFFFF;
				}

			 //	Get next part
				m_uLastPartAsked = m_pReqPartFile->GetNextRequiredPart(this);

			//	More parts ?
				if (m_uLastPartAsked != 0xFFFF)
				{
					dwCount = m_pReqPartFile->GetGapsInPart(m_uLastPartAsked, ppRequestedBlocksToAdd, dwSavedCount);
				}
			}

			for (uint32 i = 0; i < dwCount; i++)
				m_downloadBlocksList.AddTail(ppRequestedBlocksToAdd[i]);

			delete[] ppRequestedBlocksToAdd;
		}
	}

//	Why are unfinished blocks requested again, not just new ones?
	while (m_pendingBlocksList.GetCount() < 3 && !m_downloadBlocksList.IsEmpty())
	{
		Pending_Block_Struct		*pPendingBlock = new Pending_Block_Struct;

		pPendingBlock->m_pRequestedBlock = m_downloadBlocksList.RemoveHead();
		pPendingBlock->pZStream = NULL;
		pPendingBlock->pbyteRxDate = NULL;
		pPendingBlock->dwReceived = 0;
		pPendingBlock->iZStreamErr = 0;
		m_pendingBlocksList.AddTail(pPendingBlock);
	}

	if (m_pendingBlocksList.IsEmpty())
	{
		EnumDLQState eOldDowloadState = GetDownloadState();

		SendCancelTransfer();

		if ((GetTickCount() - GetLastAskedTime()) > FILEREASKTIME)
			AskForDownload();

		if (eOldDowloadState == GetDownloadState())
			SetDownloadState(DS_NONEEDEDPARTS);

		return;
	}

	Packet		*pPacket = new Packet(OP_REQUESTPARTS, 40);	// 40 = 16+(3*4)+(3*4)

	CMemFile		packetStream(reinterpret_cast<BYTE*>(pPacket->m_pcBuffer), 40);

	packetStream.Write(m_pReqPartFile->GetFileHash(), 16);

	POSITION		pos = m_pendingBlocksList.GetHeadPosition();
	uint32			adwStartEnd[6];

	Requested_Block_Struct	*pRequestedBlock;

	for (uint32 i = 0; i < 3; i++)
	{
		if (pos != NULL)
		{
			pRequestedBlock = m_pendingBlocksList.GetNext(pos)->m_pRequestedBlock;
			adwStartEnd[i] = pRequestedBlock->m_dwStartOffset;
			adwStartEnd[i + ELEMENT_COUNT(adwStartEnd) / 2] = pRequestedBlock->m_dwEndOffset + 1;
		}
		else
		{
			adwStartEnd[i] = 0;
			adwStartEnd[i + ELEMENT_COUNT(adwStartEnd) / 2] = 0;
		}
	}
	packetStream.Write(adwStartEnd, sizeof(adwStartEnd));

	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

	m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* Barry - Originally this only wrote to disk when a full 180k block
           had been received from a client, and only asked for data in
           180k blocks.

           This meant that on average 90k was lost for every connection
           to a client data source. That is a lot of wasted data.

           To reduce the lost data, packets are now written to a buffer
           and flushed to disk regularly regardless of size downloaded.
           This doesn't include compressed packets.

           Data is also requested only where gaps are, not in 180k blocks.
           The requests will still not exceed 180k, but may be smaller to
           fill a gap.
		Return:
			!= 0 - this client should be deleted by caller, else 0.
*/
int CUpDownClient::ProcessBlockPacket(char *pcPacketBuf, uint32 dwPacketSize, bool bPacked/*=false*/)
{
	EMULE_TRY

	const int		HEADER_SIZE = 24;

//	Update timer (eklmn: don't remove this value cause it use to timeout check)
	m_dwLastBlockReceived = ::GetTickCount();

//	Read data from packet
	CSafeMemFile		packetStream(reinterpret_cast<BYTE*>(pcPacketBuf), dwPacketSize);

	uchar		fileHash[16];

	packetStream.Read(fileHash, sizeof(fileHash));				//<HASH>

//	Check that this data is for the correct file
	if (m_pReqPartFile == NULL || md4cmp(fileHash, m_pReqPartFile->GetFileHash()))
	{
		throw CString(_T("wrong fileID sent, file: ") + ((m_pReqPartFile != NULL) ? m_pReqPartFile->GetFileName() : _T("NULL")));
	}

//	Find the start & end positions, and size of this chunk of data
	uint32			dwStartPos = 0;
	uint32			dwEndPos = 0;
	uint32			dwBlockSize = 0;
	const uint32 	dwDataTransferdInPacket = dwPacketSize - HEADER_SIZE;

	packetStream.Read(&dwStartPos, sizeof(uint32));				//<BYTEOFFSET:DWORD> start position

	if (bPacked)
	{
		packetStream.Read(&dwBlockSize, sizeof(uint32));		//<BYTELENGTH:DWORD> packed block size
	}
	else
	{
		packetStream.Read(&dwEndPos, sizeof(uint32));			//<BYTEOFFSET:DWORD> end position

	//	Check that packet size matches the declared packetStream packet + header size (24)
		if (dwDataTransferdInPacket != (dwEndPos - dwStartPos))
		{
			throw CString(_T("corrupted or invalid data block received"));
		}
	}

//	Protect from 0 data size, because it spoils statistics
	if (dwDataTransferdInPacket == 0)
	{
		return 0;
	}

//	Extended statistics information based on which client and remote port sent this data.
//	The new function adds the bytes to the grand total as well as the given client/port.
//	bFromPF is not relevant to downloaded data.  It is purely an uploads statistic.
	g_eMuleApp.m_pGlobPrefs->Add2SessionTransferData(GetClientSoft(), GetUserPort(), false, TRANSFERTYPE_DOWNLOAD, dwDataTransferdInPacket, false, m_pReqPartFile->GetPriority());

	m_dwTransferredDown += dwDataTransferdInPacket;

	m_dwTransferredInLastPeriod += dwDataTransferdInPacket;

//	Increment amount of data transferred for a file
	m_pReqPartFile->AddRxAmount(dwDataTransferdInPacket);

//	Loop through to find the reserved block that this is within
	Pending_Block_Struct	*cur_block;
	Requested_Block_Struct	*pReqBlk;

	POSITION pos1, pos2;

	for (pos1 = m_pendingBlocksList.GetHeadPosition(); (pos2 = pos1) != NULL; )
	{
		cur_block = m_pendingBlocksList.GetNext(pos1);
		pReqBlk = cur_block->m_pRequestedBlock;

		if ((pReqBlk->m_dwStartOffset <= dwStartPos) && (pReqBlk->m_dwEndOffset >= dwStartPos))
		{
		//	Found reserved block

			m_pCredits->AddDownloaded(dwDataTransferdInPacket, this->GetIP());

			if (cur_block->iZStreamErr != 0)
			{
			//	Increment file corruption statistics
				m_pReqPartFile->AddRxCorruptedAmount(dwDataTransferdInPacket);
			//	All data of corrupted block was received, delete the block to allow its reception on retransmission
				if ((cur_block->dwReceived += dwDataTransferdInPacket) >= dwBlockSize)
				{
				//	zstream is already deleted at that moment
					delete pReqBlk;
					delete cur_block;
					m_pendingBlocksList.RemoveAt(pos2);
				}
				break;
			}

		//	Remember this start pos, used to draw part downloading in list
			m_dwLastBlockOffset = dwStartPos;

		//	Occasionally packets are duplicated, no point writing it twice
		//	This will be 0 in these cases, or the length written otherwise
			uint32 lenWritten = 0;

		//	Handle differently depending on whether bPacked or not
			if (!bPacked)
			{
			//	Move end back one, should be inclusive
				dwEndPos--;
				cur_block->dwReceived += dwDataTransferdInPacket;

			//	Remove Hybrid, as it can generate such condition in rare cases
				if ( g_eMuleApp.m_pGlobPrefs->IsFakeRxDataFilterEnabled() &&
					(dwDataTransferdInPacket < 64) && (GetClientSoft() != SO_EDONKEYHYBRID) &&
					((pReqBlk->m_dwEndOffset - pReqBlk->m_dwStartOffset) > cur_block->dwReceived) )
				{
				//	Increment file corruption statistics
					m_pReqPartFile->AddRxCorruptedAmount(dwDataTransferdInPacket);
				//	Filter clients sending incorrect packets
					g_eMuleApp.m_pIPFilter->AddTemporaryBannedIP(GetIP());
				//	If we're logging countermeasures...
					if (!g_eMuleApp.m_pGlobPrefs->IsCMNotLog())
						AddDebugLogLine( RGB_LOG_DIMMED_TXT _T("Client '%s' (%s) (IP: %s) added to filtered clients due to bad data"),
											GetUserName(), GetClientNameAndVersionString(), GetFullIP() );
					return -1;	/*request deletion of this client by caller*/
				}

			//	Write to disk (will be buffered in part file class)
				lenWritten = m_pReqPartFile->WriteToBuffer( 0,
				             (BYTE *) (pcPacketBuf + HEADER_SIZE), dwStartPos, dwEndPos,
										 (dwEndPos >= pReqBlk->m_dwEndOffset) ? PF_WR2BUF_FL_ENDOFBLOCK : 0 );
#ifdef _DEBUG
				if (lenWritten == 0)
					AddDebugLogLine(RGB_LOG_WARNING_TXT _T("Client '%s' (%s)"), GetUserName(), GetClientNameAndVersionString());
#endif
			}
			else // Packed
			{
				uint32 lenUnzipped;

				cur_block->dwReceived += dwDataTransferdInPacket;

			//	Try to unzip the pcPacketBuf
				int result = unzip(cur_block, (BYTE *)(pcPacketBuf + HEADER_SIZE), dwDataTransferdInPacket, &lenUnzipped);

				if (result == Z_OK)
				{
				//	Write any unzipped packetStream to disk
					if (lenUnzipped != 0)
					{
					//	Use the current start and end positions for the uncompressed packetStream
						dwStartPos = pReqBlk->m_dwStartOffset;
						dwEndPos = dwStartPos + lenUnzipped - 1;
					//	Write decompressed data to the file
						lenWritten = m_pReqPartFile->WriteToBuffer(
							lenUnzipped - cur_block->dwReceived, cur_block->pbyteRxDate,
							dwStartPos, dwEndPos, PF_WR2BUF_FL_FREEBUFFER | PF_WR2BUF_FL_ENDOFBLOCK );
					//	The buffer will be freed while flushing to the file
						cur_block->pbyteRxDate = NULL;
#ifdef _DEBUG
						if (lenWritten == 0)
							AddDebugLogLine(RGB_LOG_WARNING_TXT _T("Client '%s' (%s)"), GetUserName(), GetClientNameAndVersionString());
#endif
					//	Imitate end of block, if someone will send less data or more than one compression stream
						dwEndPos = pReqBlk->m_dwEndOffset;
					}
				}
				else
				{
					cur_block->iZStreamErr = result;
					AddDebugLogLine(RGB_LOG_WARNING_TXT _T("Corrupted compressed packet for %s received (error %i)"), m_pReqPartFile->GetFileName(), result);
					m_pReqPartFile->RemoveBlockFromList(pReqBlk->m_dwStartOffset, pReqBlk->m_dwEndOffset);

				//	Release resources in case of decompression error
					if (cur_block->pZStream != NULL)
					{
						inflateEnd(cur_block->pZStream);
						delete cur_block->pZStream;
						cur_block->pZStream = NULL;
					}
					delete [] cur_block->pbyteRxDate;
					cur_block->pbyteRxDate = NULL;

				//	Increment file corruption statistics
					m_pReqPartFile->AddRxCorruptedAmount(cur_block->dwReceived);
				//	All data of corrupted block was received, delete the block to allow its reception on retransmission
					if (cur_block->dwReceived >= dwBlockSize)
					{
						delete pReqBlk;
						delete cur_block;
						m_pendingBlocksList.RemoveAt(pos2);
					}
					break;
				}
			}

		//	Check result of WriteToBuffer, if we save some bytes then
		//	this DL session can be counted as successful
			if (lenWritten > 0)
			{
			//	For determining whether the current download session was a success or not
				m_bTransferredDownMini = true;
			}

		//	We need to check "END of the block" condition every time, because if a remote client
		//	sends us data which was already written we will not delete the block & packet

			if (dwEndPos >= pReqBlk->m_dwEndOffset)
			{
				m_pReqPartFile->RemoveBlockFromList(pReqBlk->m_dwStartOffset, pReqBlk->m_dwEndOffset);
				delete pReqBlk;

			//	That should never happen
				if (cur_block->pZStream != NULL)
				{
					inflateEnd(cur_block->pZStream);
					delete cur_block->pZStream;
				}

				delete cur_block;
				m_pendingBlocksList.RemoveAt(pos2);

			//	Request next block
				SendBlockRequests();
			}

		//	Stop looping and exit method
			break;
		}
	}

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CUpDownClient::unzip(Pending_Block_Struct *pPendingBlk, BYTE *pbyteZipped, uint32 dwLenZipped, uint32 *pdwLenUnzipped)
{
	int iErr = Z_MEM_ERROR;

	EMULE_TRY

	z_stream *zS = pPendingBlk->pZStream;

//	Is this the first time this block has been unzipped
	if (zS == NULL)
	{
	//	Allocate stream
		zS = pPendingBlk->pZStream = new z_stream;

	//	Initialize stream values
		zS->zalloc = (alloc_func)0;
		zS->zfree = (free_func)0;

	//	Allocate and initialize output buffer
		zS->avail_out = pPendingBlk->m_pRequestedBlock->m_dwEndOffset - pPendingBlk->m_pRequestedBlock->m_dwStartOffset + 1;
		zS->next_out = pPendingBlk->pbyteRxDate = new BYTE[zS->avail_out];

	//	Initialise the z_stream
		iErr = inflateInit(zS);

		if (iErr != Z_OK)
			return iErr;
	}

//	Initialize input buffer
	zS->next_in = pbyteZipped;
	zS->avail_in = dwLenZipped;

//	Unzip the data
	iErr = inflate(zS, Z_SYNC_FLUSH);

	*pdwLenUnzipped = 0;
	if (iErr == Z_OK)
	{
		if (zS->avail_out == 0)
		{
		//	Output buffer is full, more than required data was sent or error occurred
			iErr = Z_BUF_ERROR;
		}
	}
//	Is zip finished reading all currently available input and writing all generated output
	else if (iErr == Z_STREAM_END)
	{
	//	Finish up
		inflateEnd(zS);
		pPendingBlk->pZStream = NULL;

		iErr = Z_OK;
	//	Got the good result, return unzipped amount
		*pdwLenUnzipped = zS->total_out;
		delete zS;
	}

	EMULE_CATCH

	return iErr;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CUpDownClient::DownloadProcess(uint32 dwReduceDownload /*in percent*/)
{
	EMULE_TRY

	uint32	dwCurTick = GetTickCount();

	if ((dwCurTick - m_dwLastBlockReceived) > DOWNLOADTIMEOUT)
	{
		SetDownloadState(DS_ONQUEUE);
		SendCancelTransfer();
	}
	else
	{
		uint32 time_diff_ms = dwCurTick - m_dwLastDataRateCalculationTime;
		uint32 dwLastTransferred = m_dwTransferredInLastPeriod;

	//	Update time & reset data transferred over last period
		m_dwTransferredInLastPeriod = 0;
		m_dwLastDataRateCalculationTime = dwCurTick;

	//	Update average sum
		m_dwClientSumDLDataRateOverLastNMeasurements += dwLastTransferred;
		m_dwClientSumDLTickOverLastNMeasurements += time_diff_ms;
	//	Update lists
		m_averageDLDataRateList.push_front(dwLastTransferred);
		m_averageDLTickList.push_front(time_diff_ms);

	//	Check lists size
		while (m_averageDLTickList.size() > 500)
		{
			m_dwClientSumDLDataRateOverLastNMeasurements -= m_averageDLDataRateList.back();
			m_averageDLDataRateList.pop_back();
			m_dwClientSumDLTickOverLastNMeasurements -= m_averageDLTickList.back();
			m_averageDLTickList.pop_back();
		}

		uint32 dwAverageTransferTimePeriod;

		if (m_dwClientSumDLTickOverLastNMeasurements != 0)
		{
		//	Calculate average time period between measuremets (ms)
			dwAverageTransferTimePeriod = m_dwClientSumDLTickOverLastNMeasurements / m_averageDLTickList.size();

		//	Calculate a DL rate (b/s)
			m_dwDownDataRate = (1000 * static_cast<uint64>(m_dwClientSumDLDataRateOverLastNMeasurements)) / m_dwClientSumDLTickOverLastNMeasurements;
		}
		else
		{
			dwAverageTransferTimePeriod = 100;
		//	Calculate a DL rate (b/s) assuming 100 ms interval
			m_dwDownDataRate = 10 * m_dwClientSumDLDataRateOverLastNMeasurements;
		}

	//	now based on measured speed we can control the bandwidth
		if (dwReduceDownload)
		{
			m_bLimitlessDL = false;
		//	the limit for next period is calculated as required DL per millisecond multiplicated by the time period.
		//		required DL (Byte/s) = percent * measured DL (Byte/s)
		//		required DL (Byte/ms) = required DL (Byte/s) / 1000
		//		Limit = required DL (Byte/ms) * sample period (ms)
		//	The previous formula assumed 100ms period between two speed calculations
		//		uint32 limit = dwReduceDownload * dwCurrentDataRate / 1000;
		//			where dwReduceDownload = percent * 100
		//	Since we are calculating the average period value, then we gonna use this value
			uint32 dwUpperLimit = (dwReduceDownload * m_dwDownDataRate * dwAverageTransferTimePeriod)/100000;
			uint32 dwLowerLimit = (200 * dwAverageTransferTimePeriod)/1000;

			if (dwUpperLimit < 1000 && dwReduceDownload == 200)
				dwUpperLimit += 1000;
			else if (dwUpperLimit < dwLowerLimit)
				dwUpperLimit = dwLowerLimit;
#ifdef OLD_SOCKETS_ENABLED
			m_pRequestSocket->SetDownloadLimit(dwUpperLimit);
#endif OLD_SOCKETS_ENABLED
		}
#ifdef OLD_SOCKETS_ENABLED
		else if (!m_bLimitlessDL)
		{
			m_pRequestSocket->DisableDownloadLimit();
			m_bLimitlessDL = true;
		}
#endif OLD_SOCKETS_ENABLED
	}

	UpdateDisplayedInfo();

	return m_dwDownDataRate;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint16 CUpDownClient::GetAvailablePartCount()
{
	EMULE_TRY

	uint16 result = 0;

	for (int i = 0;i < m_uPartCount;i++)
	{
		if (IsPartAvailable(i))
			result++;
	}

	return result;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetRemoteQueueRank(uint16 nr, bool bUpdateGUI /*= true*/)
{
	EMULE_TRY

	m_iDeltaQueueRank = static_cast<int>(nr) - static_cast<int>(m_uRemoteQueueRank);
	if (m_uRemoteQueueRank != nr)
	{
		//	if client sent us an opcode with non-0 rank, that it means we are on his WQ
		//	so if we had 0 rank before just enter into his WQ
		if (m_uRemoteQueueRank == 0)
			StartDLQueueWaitTimer();

		m_uRemoteQueueRank = nr;
		if (bUpdateGUI)
			UpdateDisplayedInfo();
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// A4AF switching
bool CUpDownClient::SwapToAnotherFile(CPartFile* SwapTo, bool bOneWaySwap /*false*/, bool bIgnoreReasktime /*false*/)
{
	EMULE_TRY

//	Check if we are downloading something from client
	if (GetDownloadState() == DS_DOWNLOADING)
		return false;

	POSITION		SwapToPos = 0;
	POSITION		cur_FilePos;
	CPartFile	   *pCurFile;
	uint32			dwLastNNSCheckedTime = ::GetTickCount();
	const uint16	uPriorityFactorsArray[3] = {1,2,3};
	uint16			uA4AFRating;
	uint16			uBestA4AFRating = 0;
	bool			bIsFileReaskAllowed = true;
	bool			bNNS  = false;

//	Check: does client have another file?
	if (!m_otherRequestsList.IsEmpty())
	{
	//	Check: target file is specified? if is not, then find another one...
		if (!SwapTo)
		{
		//	Try to find a file in the same category
			CPartFile	*pTempPartFile;
			POSITION	pos;
			bool		bFoundInSameCat = false;

			for (pos = m_otherRequestsList.GetHeadPosition(); pos != NULL;)
			{
				pTempPartFile = m_otherRequestsList.GetNext(pos);
				if (CCat::FileBelongsToGivenCat(pTempPartFile, m_pReqPartFile->GetCatID()))
				{
					bFoundInSameCat = true;
					break;
				}
			}

		//	Find a file to switch to (now taking in account alphabetic order inside categories
		//	Do not change the way it's done (for loop and pCurFile assignment) because, I don't know why,
		//	it produces unexpected results. :(
			for (cur_FilePos = m_otherRequestsList.GetHeadPosition();cur_FilePos != 0;m_otherRequestsList.GetNext(cur_FilePos))
			{
				pCurFile = m_otherRequestsList.GetAt(cur_FilePos);

			//	Get first high priority files
				if ( pCurFile != m_pReqPartFile && g_eMuleApp.m_pDownloadQueue->IsInDLQueue(pCurFile)
				     && (pCurFile->GetStatus(false) == PS_READY || pCurFile->GetStatus(false) == PS_EMPTY) )
				{
				//	Check: is new file in NNS-Map?
					bNNS = m_otherNoNeededMap.Lookup(pCurFile, dwLastNNSCheckedTime);

				//	Calculate A4AF rating
					if (bNNS)
					{
						uA4AFRating = uPriorityFactorsArray[pCurFile->GetPriority()]
							+ static_cast<uint16>((::GetTickCount() - dwLastNNSCheckedTime)/FILEREASKTIME);
						bIsFileReaskAllowed = (::GetTickCount() - dwLastNNSCheckedTime) > FILEREASKTIME;
					}
					else
					{
						uA4AFRating = uPriorityFactorsArray[pCurFile->GetPriority()] * 3;
						bIsFileReaskAllowed = true;
					}

					if ((uA4AFRating >= uBestA4AFRating) && (bIsFileReaskAllowed || bIgnoreReasktime))
					{
						if(!SwapTo || !bFoundInSameCat || (SwapTo->GetFileName() > pCurFile->GetFileName()))
						{
							SwapTo = pCurFile;
							SwapToPos = cur_FilePos;
							uBestA4AFRating = uA4AFRating;
						}
					}
				}
			}
		}
		else if (SwapTo->GetStatus(false) == PS_READY || SwapTo->GetStatus(false) == PS_EMPTY)
		{
			bNNS = m_otherNoNeededMap.Lookup(SwapTo, dwLastNNSCheckedTime);

			if (bNNS)
			{
				bIsFileReaskAllowed = (::GetTickCount() - dwLastNNSCheckedTime) > FILEREASKTIME;
			}
			if (!bNNS || bIsFileReaskAllowed || bIgnoreReasktime)
			{
				SwapToPos = m_otherRequestsList.Find(SwapTo);
			}
		}

	//	Now let's swap the file to another
		if (SwapTo != NULL && SwapToPos != 0)
		{
		//	To prevent loop swaping remember NNS-Part & time when it was switched
			if (bOneWaySwap)
				m_otherNoNeededMap.RemoveKey(m_pReqPartFile);
			else if ( m_pReqPartFile != NULL /* && GetDownloadState() == DS_NONEEDEDPARTS */ )
			{
				m_otherNoNeededMap[m_pReqPartFile] = ::GetTickCount();
			}

		//	Switch state to NONE
			SetDownloadState(DS_NONE);

			if (m_pbytePartStatuses)
			{
				delete[] m_pbytePartStatuses;
				m_pbytePartStatuses = 0;
			}

			m_uPartCount = 0;
			m_dwLastAskedTime = 0;
			m_uLastPartAsked = 0xFFFF;

		//	Add to past comment
			if (m_pReqPartFile != NULL)
			{
				m_pReqPartFile->AddPastComment(this);
				m_eRating = PF_RATING_NONE;
				m_strComment = _T("");
			//	Find current file position in sourceslots
				cur_FilePos = m_pReqPartFile->m_srcLists[m_iSourcesSlot].Find(this);

 			//	Remove current file
				if (cur_FilePos)
					m_pReqPartFile->m_srcLists[m_iSourcesSlot].RemoveAt(cur_FilePos);

			//	Add them in A4AF list
				if (!bOneWaySwap && !m_pReqPartFile->m_A4AFsrcList.Find(this))
					m_pReqPartFile->m_A4AFsrcList.AddTail(this);

			//	Add old file to m_otherRequestsList
				if (!bOneWaySwap && !m_otherRequestsList.Find(m_pReqPartFile))
					m_otherRequestsList.AddTail(m_pReqPartFile);

			//	Let's update current information for the removed file
				m_pReqPartFile->NewSrcPartsInfo();

			//	Update download list (old file update)
				m_pReqPartFile->UpdateDisplayedInfo();

			//	Reset before delete m_pbytePartStatuses
				m_pReqPartFile = NULL;
			}

		//	Actual switch
			m_pReqPartFile = SwapTo;

		//	Remove new file from OtherRequests list
			m_otherRequestsList.RemoveAt(SwapToPos);

		//	Remove current source from A4AF list in new file
			POSITION	pos_A4AF = SwapTo->m_A4AFsrcList.Find(this);

			if (pos_A4AF != NULL)
				SwapTo->m_A4AFsrcList.RemoveAt(pos_A4AF);

		//	To prevent loop swaping remove a choosen part from NNS-Map
			m_otherNoNeededMap.RemoveKey(SwapTo);

		//	Add to souceslots
			if (!SwapTo->m_srcLists[m_iSourcesSlot].Find(this))
				SwapTo->m_srcLists[m_iSourcesSlot].AddTail(this);

		//	Remove from past comment
			SwapTo->RemovePastComment(this);

		//	Update info for target file
			SwapTo->NewSrcPartsInfo();

		//	Update download list (new file update)
			SwapTo->UpdateDisplayedInfo();

		//	Now update the client item
			UpdateDisplayedInfo();

			return (m_pReqPartFile != NULL);
		}
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::UDPReaskACK(uint16 nNewQR)
{
	EMULE_TRY

	m_bUDPPending = false;
	SetRemoteQueueRank(nNewQR);
	m_dwLastAskedTime = ::GetTickCount();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::UDPReaskFNF()
{
	EMULE_TRY

	m_bUDPPending = false;

	if (GetDownloadState() != DS_DOWNLOADING)
	{
	//	Try to swap the source to another file before removing it from DL-queue
		if (!SwapToAnotherFile(NULL, true))
		{
#ifdef OLD_SOCKETS_ENABLED
	//	if client have socket or was not removed indirectly over disconnection, remove him from DLQ
		if (!m_pRequestSocket && !Disconnected())
			g_eMuleApp.m_pDownloadQueue->RemoveSource(this, true);
#endif //OLD_SOCKETS_ENABLED
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::UDPReaskForDownload()
{
	EMULE_TRY

	ASSERT (m_pReqPartFile);

	if (!m_pReqPartFile || m_bUDPPending)
		return;

//	Disable UDP when connecting to a PROXY
	ProxySettings proxy = g_eMuleApp.m_pGlobPrefs->GetProxy();

#ifdef OLD_SOCKETS_ENABLED

	if ( m_byteEmuleVersion >= 0x23
		&& m_uUDPPort != 0
		&& g_eMuleApp.m_pGlobPrefs->GetUDPPort() != 0
		&& (!g_eMuleApp.m_pServerConnect->IsLowID() && !HasLowID() )
		&& !(m_pRequestSocket && m_pRequestSocket->IsConnected())
		&& (!proxy.m_bUseProxy) )
	{

	//	Don't use UDP to ask for sources
		if (IsSourceRequestAllowed())
			return;

		m_bUDPPending = true;

		CSafeMemFile	packetStream(128);

		packetStream.Write(m_pReqPartFile->GetFileHash(), 16);
		if (GetUDPVersion() > 3)
		{
			if (m_pReqPartFile->IsPartFile())
				((CPartFile*)m_pReqPartFile)->WritePartStatus(&packetStream);
			else
			{
				uint16	uNull = 0;

				packetStream.Write(&uNull, 2);
			}
		}
		if (GetUDPVersion() > 2)
		{
			uint16	uCompleteSourcesCount = m_pReqPartFile->GetCompleteSourcesCount();
			packetStream.Write(&uCompleteSourcesCount, 2);
		}

		Packet		*pPacket = new Packet(&packetStream, OP_EMULEPROT);

		pPacket->m_eOpcode = OP_REASKFILEPING;

		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);

		g_eMuleApp.m_pClientUDPSocket->SendPacket(pPacket, GetIP(), GetUDPPort());

		AddAskedCountDown();
	}

#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Sets the vector of bool 'parts' to show the pending status of each part.
void CUpDownClient::ShowDownloadingParts(_Bvector& parts)
{
	EMULE_TRY

	parts.resize(m_uPartCount);

	for (POSITION pos = m_pendingBlocksList.GetHeadPosition(); pos != NULL; )
		parts[m_pendingBlocksList.GetNext(pos)->m_pRequestedBlock->m_dwStartOffset / PARTSIZE] = true;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::UpdateDisplayedInfo()
{
	g_eMuleApp.m_pDownloadList->UpdateSource(this);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::ClearPendingBlocksList(void)
{
	EMULE_TRY

	Pending_Block_Struct*		pPendingBlock;

	while (!m_pendingBlocksList.IsEmpty() )
	{
		pPendingBlock = m_pendingBlocksList.RemoveHead();

		if (pPendingBlock != NULL)
		{
			CPartFile *pReqFile = g_eMuleApp.m_pDownloadQueue->GetFileByID(pPendingBlock->m_pRequestedBlock->m_fileHash);

			if (pPendingBlock->pZStream != NULL)
			{
				uint32	dwTotalOut;

				if ((dwTotalOut = pPendingBlock->pZStream->total_out) != 0)
				{
				//	Some data was decompressed

				//	Force file write of partly received compressed block
				//	The data buffer will be freed while flushing to the file
					pReqFile->WriteToBuffer( dwTotalOut - pPendingBlock->pZStream->total_in,
						pPendingBlock->pbyteRxDate, pPendingBlock->m_pRequestedBlock->m_dwStartOffset,
						pPendingBlock->m_pRequestedBlock->m_dwStartOffset + dwTotalOut - 1,
						PF_WR2BUF_FL_FREEBUFFER | PF_WR2BUF_FL_ENDOFBLOCK );
				}
				inflateEnd(pPendingBlock->pZStream);
				delete pPendingBlock->pZStream;
			}
			pReqFile->RemoveBlockFromList(pPendingBlock->m_pRequestedBlock->m_dwStartOffset,
										pPendingBlock->m_pRequestedBlock->m_dwEndOffset);
			delete pPendingBlock->m_pRequestedBlock;
			delete pPendingBlock;
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	If downloadBlocks are existing free block reserved in pPartFile
void CUpDownClient::ClearDownloadBlocksList(void)
{
	while (!m_downloadBlocksList.IsEmpty())
	{
		Requested_Block_Struct* pDLBlock =  m_downloadBlocksList.RemoveHead();
		if (pDLBlock)
		{
			CPartFile* pReqFile = g_eMuleApp.m_pDownloadQueue->GetFileByID(pDLBlock->m_fileHash);

			pReqFile->RemoveBlockFromList(pDLBlock->m_dwStartOffset, pDLBlock->m_dwEndOffset);
			delete pDLBlock;
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendStartupLoadRequest()
{
	Packet	*pPacket = NULL;

	try
	{
		if (m_pReqPartFile != NULL && m_pRequestSocket != NULL)
		{
			Packet 	*pPacket = new Packet(OP_STARTUPLOADREQ, 16);

			md4cpy(pPacket->m_pcBuffer, m_pReqPartFile->GetFileHash());
			g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
			m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED

		//	don't switch the status in case of NNS if we already have the part status array
			if (GetDownloadState() != DS_NONEEDEDPARTS || m_pbytePartStatuses == NULL)
				SetDownloadState(DS_ONQUEUE);
			m_dwLastAskedTime = ::GetTickCount();
		}
	}
	catch (CException * error)
	{
		safe_delete(pPacket);
		g_eMuleApp.m_pdlgEmule->AddBugReport(__FUNCTION__,__FILE__, __LINE__,GetErrorMessage(error));
		error->Delete();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendHashsetRequest()
{
	Packet	*pPacket = NULL;

	try
	{
		if (m_pReqPartFile != NULL)
		{
			pPacket = new Packet(OP_HASHSETREQUEST, 16);

			md4cpy(pPacket->m_pcBuffer, m_pReqPartFile->GetFileHash());
			g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
			m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED

			SetDownloadState(DS_REQHASHSET);
			m_pReqPartFile->m_bHashSetNeeded = false;
			m_fRequestingHashSet = 1;

			AddDebugLogLine(_T("Sent hashset request with filehash %s to client '%s' (%s)"), HashToString(m_pReqPartFile->GetFileHash()), GetUserName(), GetClientNameAndVersionString());
		}
	}
	catch (CException * error)
	{
		safe_delete(pPacket);
		g_eMuleApp.m_pdlgEmule->AddBugReport(__FUNCTION__,__FILE__, __LINE__,GetErrorMessage(error));
		error->Delete();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// return the remainig download time for the current chunk in second
uint32 CUpDownClient::GetRemainingTimeForCurrentPart()
{
	uint32 dwSecond = 0;

	if (m_pReqPartFile != NULL
		&& m_eDownloadState ==  DS_DOWNLOADING
		&& m_uLastPartAsked != 0xFFFF 
		&& m_dwDownDataRate != 0)
	{
		uint32 dwPartSize = m_pReqPartFile->GetPartLeftToDLSize(m_uLastPartAsked);
		
		dwSecond = dwPartSize / m_dwDownDataRate;
	}

	return dwSecond;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// return the remainig download size for the current chunk in bytes
uint32 CUpDownClient::GetRemainingSizeForCurrentPart()
{
	uint32 dwSize = 0;

	if (m_pReqPartFile != NULL
		&& m_eDownloadState ==  DS_DOWNLOADING
		&& m_uLastPartAsked != 0xFFFF)
	{
		dwSize = m_pReqPartFile->GetPartLeftToDLSize(m_uLastPartAsked);
	}

	return dwSize;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendCancelTransfer()
{
	Packet	*pPacket = NULL;

	try
	{
		if (m_pRequestSocket)
		{
			pPacket = new Packet(OP_CANCELTRANSFER);

			g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);

#ifdef OLD_SOCKETS_ENABLED
			m_pRequestSocket->SendPacket(pPacket, true, true);
		}
#endif //OLD_SOCKETS_ENABLED	
	}
	catch (CException * error)
	{
		safe_delete(pPacket);
		g_eMuleApp.m_pdlgEmule->AddBugReport(__FUNCTION__,__FILE__, __LINE__,GetErrorMessage(error));
		error->Delete();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
