//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "DownloadQueue.h"
#include "updownclient.h"
#include "partfile.h"
#include "ed2klink.h"
#include "otherfunctions.h"
#include "Smtp.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

BEGIN_MESSAGE_MAP(CHostnameSourceWnd, CWnd)
	ON_MESSAGE(TM_SOURCEHOSTNAMERESOLVED, OnSourceHostnameResolved)
	ON_WM_TIMER()
END_MESSAGE_MAP()

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CHostnameSourceWnd::CHostnameSourceWnd()
{
	m_pOwner = NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
LRESULT CHostnameSourceWnd::OnSourceHostnameResolved(WPARAM wParam, LPARAM lParam)
{
	if (m_pOwner)
		m_pOwner->SourceHostnameResolved(wParam, lParam);

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CDownloadQueue::CDownloadQueue(CPreferences *in_pPrefs, CSharedFileList *in_pSharedFileList)
{
	m_pPrefs = in_pPrefs;
	m_pSharedFileList = in_pSharedFileList;
	m_iDataRate = 0;
	pCurUDPServer = NULL;
	m_dwLastUDPSearchTime = 0;
	m_dwLastUDPStatTime = 0;
	m_pLastUDPSearchedFile = NULL;
	m_iAverageDataRate = 0;
	m_iUDCounter = 0;
	m_dwSumDownDataOverheadInDeque = 0;
	m_nDownDataRateMSOverhead = 0;
	m_nDownDataRateOverhead = 0;
	m_nDownDataOverheadSourceExchange = 0;
	m_nDownDataOverheadFileRequest = 0;
	m_nDownDataOverheadOther = 0;
	m_nDownDataOverheadServer = 0;
	m_nDownDataOverheadSourceExchangePackets = 0;
	m_nDownDataOverheadFileRequestPackets = 0;
	m_nDownDataOverheadOtherPackets = 0;
	m_nDownDataOverheadServerPackets = 0;
	m_A4AF_auto_file = NULL;
	m_bIsInitialized = false;
//	We need to initialize this variable to prevent extra screen refreshing during start
	m_dwLastRefreshedDLDisplayTime = ::GetTickCount();
	m_dwLastTCPSourcesRequestTime = 0;
	m_pLastTCPSrcReqServer = NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CDownloadQueue::~CDownloadQueue()
{
	EMULE_TRY

//--- xrmb:keepPartFileStats ---
	SaveAllPartFileStats();
//--- :xrmb ---

	while (!m_partFileList.IsEmpty())
	{
		CPartFile* pPartFile = dynamic_cast<CPartFile*>(m_partFileList.RemoveHead());

		if (pPartFile)
			delete pPartFile;
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::AddPartFilesToShare()
{
	EMULE_TRY

	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL;)
	{
		CPartFile	* pPartFile = dynamic_cast<CPartFile*>(m_partFileList.GetNext(pos));

		if (pPartFile && pPartFile->GetStatus(true) == PS_READY)
			m_pSharedFileList->SafeAddKnownFile(pPartFile, true);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::CompDownDataRateOverhead()
{
	EMULE_TRY

//	update the list & sum
	m_DownDataOverheadDeque.push_back(m_nDownDataRateMSOverhead);
	m_dwSumDownDataOverheadInDeque += m_nDownDataRateMSOverhead;

//	reset the data
	m_nDownDataRateMSOverhead = 0;

//	chech deque length
	while (m_DownDataOverheadDeque.size() > 150)
	{
		m_dwSumDownDataOverheadInDeque -= m_DownDataOverheadDeque.front();
		m_DownDataOverheadDeque.pop_front();
	}

//	calculate average value
	if (m_DownDataOverheadDeque.size() > 10)
		m_nDownDataRateOverhead = 10 * m_dwSumDownDataOverheadInDeque/m_DownDataOverheadDeque.size();
	else
		m_nDownDataRateOverhead = 0;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::Init()
{
	EMULE_TRY

//	Find all part files, read & hash them if needed and store them into a list
	uint16	iCount = 0;
	CString	strTempDir = m_pPrefs->GetTempDir();
	CStringList	tmpTempDirList;	//	list elements will be deleted in list destructor

//	Make local copy to prevent long locking of list resource
	m_pPrefs->TempDirListCopy(&tmpTempDirList);
//	Add the main temporary directory to the temporary directories list for easier processing
	strTempDir += _T('\\');
	tmpTempDirList.AddHead(strTempDir);

//	For each temp dir...
	for (POSITION pos = tmpTempDirList.GetHeadPosition(); pos != NULL; )
	{
		CFileFind		ff;
		CString			strTempDirPath(tmpTempDirList.GetNext(pos));
		CString			strSearchPath = strTempDirPath;

		strSearchPath += _T("*.part.met");

	//	Check all part.met files
		bool	bEnd = !ff.FindFile(strSearchPath, 0);

	//	Until we can't find any more .met files...
		while (!bEnd)
		{
			bEnd = !ff.FindNextFile();
		//	If, for some reason, there's a .met directory, skip it
			if (ff.IsDirectory())
				continue;

			CPartFile		*pPartFile = new CPartFile();

		//	Try to load the part file. If we were successful...
		
			bool bLoadPartFileStatus = pPartFile->LoadPartFile(strTempDirPath, ff.GetFileName().GetBuffer());

		//	Try recovering part.met.bak file if load failed before
			if (!bLoadPartFileStatus)
				bLoadPartFileStatus = pPartFile->TryToRecoverPartFile(strTempDirPath, ff.GetFileName().GetBuffer());

			if (bLoadPartFileStatus)
			{
				iCount++;
			//	Check if is an A4AF auto file
				if (md4cmp(pPartFile->GetFileHash(), g_eMuleApp.m_pGlobPrefs->GetA4AFHash()) == 0)
					g_eMuleApp.m_pDownloadQueue->SetA4AFAutoFile(pPartFile);

			//	Add it to the download queue
				m_partFileList.AddTail(pPartFile);

			//	PartFiles are always shared files
				if (pPartFile->GetStatus(true) == PS_READY)
					m_pSharedFileList->SafeAddKnownFile(pPartFile);

				g_eMuleApp.m_pDownloadList->AddFile(pPartFile);

			//	Load SLS sources on PartFile loading
				m_sourcesaver.LoadSources(pPartFile);
			}
			else
			{
				delete pPartFile;
			}
		}
		ff.Close();
	}

	if (iCount == 0)
	{
		AddLogLine(false, IDS_NOPARTSFOUND);
	}
	else
	{
		AddLogLine(false, IDS_FOUNDPARTS, iCount);
		SortByPriority();
	}

//--- xrmb:keepPartFileStats ---
	m_lastPartFileStatsSave = GetTickCount();
//--- :xrmb ---

	m_wndHSCallback = NULL;
	sourceHostnameResolveRetry = 0;
	m_bIsResolving = false;
	m_bIsInitialized = true;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::AddSearchToDownload(CSearchFile *pNewSearchFile, EnumCategories eCatID /*=CAT_NONE*/, bool bPaused)
{
	EMULE_TRY

	if (pNewSearchFile->GetFileSize() == 0)
	{
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_SKIPZEROLENGTHFILE), pNewSearchFile->GetFileName());
		return ;
	}
	if (FileExists(pNewSearchFile->GetFileHash()))
		return ;

	CPartFile	*pNewPartFile = new CPartFile(pNewSearchFile, eCatID);

	if (pNewPartFile->GetStatus() == PS_ERROR)
	{
		delete pNewPartFile;
		return ;
	}
	AddDownload(pNewPartFile, bPaused);

//	If the search result is from a client ('View Files') add that client as source
	if ( pNewSearchFile->GetType() == SFT_CLIENT
	  && pNewSearchFile->GetClientID()
	  && pNewSearchFile->GetClientPort() )
	{
		CMemFile	sourcesFile;

		uint16		uSources = 1;
		sourcesFile.Write(&uSources, sizeof uSources);

		uint32		uIP = pNewSearchFile->GetClientID();
		sourcesFile.Write(&uIP, sizeof uIP);

		uint16		uPort = pNewSearchFile->GetClientPort();
		sourcesFile.Write(&uPort, sizeof uPort);

		uIP = pNewSearchFile->GetClientServerIP();
		sourcesFile.Write(&uIP, sizeof uIP);

		uPort = pNewSearchFile->GetClientServerPort();
		sourcesFile.Write(&uPort, sizeof uPort);

		sourcesFile.SeekToBegin();
		if (pNewPartFile)
		{
			pNewPartFile->AddClientSources(&sourcesFile);
		}
		else
		{
			CPartFile	*pPartFile = GetFileByID(pNewSearchFile->GetFileHash());

			if (pPartFile)
				pPartFile->AddClientSources(&sourcesFile);
		}
	}
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateCatTabTitles();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::AddFileLinkToDownload(CED2KFileLink *pLink, EnumCategories eCatID)
{
	EMULE_TRY

	CPartFile	*pNewPartFile = new CPartFile(pLink, eCatID);

	if (pNewPartFile->GetStatus() == PS_ERROR)
	{
		delete pNewPartFile;
		pNewPartFile = NULL;
	}
	else
	{
		AddDownload(pNewPartFile, g_eMuleApp.m_pGlobPrefs->StartDownloadPaused());
	}

	if (pLink->HasValidSources())
	{
		if (pNewPartFile)
		{
			if (pLink->ClientSourcesList)
			//	Add all IP sources and eliminate them from list
				pNewPartFile->AddClientSources(pLink->ClientSourcesList);
		}
		else
		{
			CPartFile	*pPartFile = GetFileByID(pLink->GetHashKey());

			if (pPartFile && pLink->ClientSourcesList)
			//	Add all IP sources and eliminate them from list
				pPartFile->AddClientSources(pLink->ClientSourcesList);
		}
	}
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateCatTabTitles();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	StartNextFile() starts a paused download preferably in the same category (takes in account priority)
void CDownloadQueue::StartNextFile(EnumCategories eCatID)
{
	EMULE_TRY

	CPartFile	*pPartFile, *pTempPartFile;
	CPartFile	*pStartFile = NULL;
	POSITION	pos;

	bool bFoundInSameCat = false;

//	Try to find a paused file in the same category
	for (pos = m_partFileList.GetHeadPosition(); pos != NULL; )
	{
		pTempPartFile = dynamic_cast<CPartFile*>(m_partFileList.GetNext(pos));
		if (CCat::FileBelongsToGivenCat(pTempPartFile, eCatID) && pTempPartFile->GetStatus() == PS_PAUSED)
		{
			bFoundInSameCat = true;
			break;
		}
	}
	for (pos = m_partFileList.GetHeadPosition(); pos != NULL; )
	{
		pPartFile = dynamic_cast<CPartFile*>(m_partFileList.GetNext(pos));

		if (pPartFile && pPartFile->GetStatus() == PS_PAUSED && ((!bFoundInSameCat && g_eMuleApp.m_pGlobPrefs->IsResumeOtherCat())
			|| (bFoundInSameCat &&  CCat::FileBelongsToGivenCat(pPartFile, eCatID))))
		{
		//	If no start file has been found or 'pPartFile' is higher priority...
		//  and now it's same priority but next in file order (ie. 1/2/3 in a series)
			if (pStartFile == NULL || pStartFile->GetPriority() < pPartFile->GetPriority()
				|| ( pStartFile->GetPriority() == pPartFile->GetPriority()
				    && pStartFile->GetFileName() > pPartFile->GetFileName() ) )
				pStartFile = pPartFile;
		}
	}

	if (pStartFile != NULL)
		pStartFile->ResumeFile();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::AddDownload(CPartFile *pNewPartFile, bool bPaused)
{
	EMULE_TRY

	m_partFileList.AddTail(pNewPartFile);
	SortByPriority();

	if (bPaused)
	{
		pNewPartFile->PauseFile();
		pNewPartFile->SetStartTimeReset(true);
	}

	g_eMuleApp.m_pDownloadList->AddFile(pNewPartFile);

	CCat		*pCat = CCat::GetCatByID(pNewPartFile->GetCatID());
	CString		strCategory = (pCat != NULL) ? pCat->GetTitle() : GetResString(IDS_CAT_UNCATEGORIZED);

	AddLogLine(true, IDS_NEWDOWNLOAD, pNewPartFile->GetFileName(), strCategory);

//	Sending message when a download is added
	CString	strMessageText;

	strMessageText.Format(GetResString(IDS_NEWDOWNLOAD), pNewPartFile->GetFileName(), strCategory);
	g_eMuleApp.m_pSMTPConnection->SendMail( strMessageText, g_eMuleApp.m_pGlobPrefs->GetUseDownloadAddNotifier(),
	                                        g_eMuleApp.m_pGlobPrefs->IsSMTPInfoEnabled() );
	g_eMuleApp.m_pdlgEmule->ShowNotifier( strMessageText, TBN_DLOAD_ADD, false,
	                                      g_eMuleApp.m_pGlobPrefs->GetUseDownloadAddNotifier() );
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CDownloadQueue::FileExists(const uchar *strFileHash)
{
	bool	bFileExists = false;

	EMULE_TRY

	if (CKnownFile *pKnownFile = m_pSharedFileList->GetFileByID(strFileHash))
	{
		if (dynamic_cast<CPartFile*>(pKnownFile) != NULL)
			AddLogLine(true, IDS_ERR_ALREADY_DOWNLOADING, pKnownFile->GetFileName());
		else
			AddLogLine(true, IDS_ERR_ALREADY_DOWNLOADED, pKnownFile->GetFileName());
		bFileExists = true;
	}
	else if (pKnownFile = GetFileByID(strFileHash))
	{
		AddLogLine(true, IDS_ERR_ALREADY_DOWNLOADING, pKnownFile->GetFileName());
		bFileExists = true;
	}

	EMULE_CATCH

	return bFileExists;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::Process()
{
	EMULE_TRY

	if (!m_bIsInitialized)
		return ;

	uint32	iDownSpeed = 0;
	uint32	dwMaxDLSpeed = m_pPrefs->GetMaxDownload();

	if ((dwMaxDLSpeed != UNLIMITED) && ((m_iDataRate > 1500) || (dwMaxDLSpeed < 30)))
	{
		uint32	dwMaxDLSpeedBPS = dwMaxDLSpeed * 1024;	// Bytes per second

	//	Don't remove upper DL limit, otherwise one DL will block all others
	//	Calculate what percentage of the max data rate the current data rate is
		iDownSpeed = (dwMaxDLSpeedBPS * 10) / (m_iDataRate + 1);
	//	Pin it between 50% and 200% of max
		if (iDownSpeed < 50)
			iDownSpeed = 50;
		else if (iDownSpeed > 200)
			iDownSpeed = 200;
	}

//	Process all the part files in the queue in decreasing order of priority
//	m_partFileList is already sorted by prio every time a file is added or deleted
//	therefore we removed all the extra loops...
	uint32	iTempDataRate = 0;

	m_iUDCounter++;

	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL; )
	{
		CPartFile	*pPartFile = dynamic_cast<CPartFile*>(m_partFileList.GetNext(pos));

		if (pPartFile)
		{
			EnumPartFileStatuses	eFileStatus = pPartFile->GetStatus();

			if ((eFileStatus == PS_READY) || (eFileStatus == PS_EMPTY))
			{
				iTempDataRate += pPartFile->Process(iDownSpeed, m_iUDCounter /*iteration*/);
			}
		}
	}

	m_iDataRate = iTempDataRate;

	DWORD	dwCurTick = ::GetTickCount();

//	Save used bandwidth for speed calculations
	if (m_averageTickList.empty() || dwCurTick - m_averageTickList.front() >= 500)	// Update no faster than every .5 sec
	{
		m_averageDataRateList.push_front(g_eMuleApp.stat_sessionReceivedBytes);

	//	Save time between each speed snapshot
		m_averageTickList.push_front(dwCurTick);

	//	Save 40 secs of data. It seems to give the most accurate result atm...
		while (dwCurTick - m_averageTickList.back() > 40000)
		{
			m_averageDataRateList.pop_back();
			m_averageTickList.pop_back();
		}
	}

//	Calculate average data rate
	if (m_averageDataRateList.size() > 1)
		m_iAverageDataRate = static_cast<uint32>((static_cast<double>(m_averageDataRateList.front() - m_averageDataRateList.back())) * 1000.0 / (m_averageTickList.front() - m_averageTickList.back()));
	else
		m_iAverageDataRate = 0;

	dwCurTick = ::GetTickCount();

//	The code of sources refresh on local server was moved here in order prevent the situation
//	where the client is blacklisted. We create one 'packet' which contains the
//	15 buffered OP_GETSOURCES ED2K packets to be sent in one TCP frame every 300 sec
#ifdef OLD_SOCKETS_ENABLED
	if (m_iUDCounter == 2 && g_eMuleApp.m_pServerConnect->IsConnected())
	{
		if (!m_dwLastTCPSourcesRequestTime || (dwCurTick - m_dwLastTCPSourcesRequestTime) > SERVER_SRC_REQ_QUANT)
		{
			int iRequestedNumber = 0;
			CSafeMemFile		packetStream(22);

			while (iRequestedNumber < 15 && !m_LocalServerSourcesReqQueue.IsEmpty())
			{
				CPartFile	   *pPartFile = m_LocalServerSourcesReqQueue.RemoveHead();

				if (pPartFile != NULL && (pPartFile->GetStatus() == PS_READY || pPartFile->GetStatus() == PS_EMPTY))
				{
					Packet	   *packet = new Packet(OP_GETSOURCES, 16);

					md4cpy(packet->m_pcBuffer, pPartFile->GetFileHash());
					packetStream.Write(packet->GetPacket(), packet->GetRealPacketSize());
					delete packet;
					iRequestedNumber++;
				}
			}

			//	Create one 'packet' which contains all buffered OP_GETSOURCES ED2K packets to be sent in one TCP frame
			//	KuSh: in the official code bFromPF arg is set to false ? isn't this packet only sent for PartFile?
			const int	iStreamSize = packetStream.GetLength();

			if (iStreamSize > 0)
			{
				Packet	*packet = new Packet(new char[iStreamSize], iStreamSize, true, false);

				packetStream.SeekToBegin();
				packetStream.Read(packet->GetPacket(), iStreamSize);
				g_eMuleApp.m_pUploadQueue->AddUpDataOverheadServer(packet->m_dwSize);
				g_eMuleApp.m_pServerConnect->SendPacket(packet, true);
			//	Update time only in case we sent something
				m_dwLastTCPSourcesRequestTime = ::GetTickCount();
			}
		}
	}
#endif //OLD_SOCKETS_ENABLED

	if (m_iUDCounter == 5)
	{
		if ((!m_dwLastUDPStatTime) || (dwCurTick - m_dwLastUDPStatTime) > UDPSERVERSTATTIME)
		{
			m_dwLastUDPStatTime = dwCurTick;
			g_eMuleApp.m_pServerList->ServerStats();
		}
	}

	if (m_iUDCounter == 10)
	{
		m_iUDCounter = 0;
		if ((!m_dwLastUDPSearchTime) || (dwCurTick - m_dwLastUDPSearchTime) > UDPSERVERREASKTIME)
		{
			SendNextUDPPacket();
		}
	}

//--- xrmb:keepPartFileStats ---
	if (m_lastPartFileStatsSave + 300000 < dwCurTick)
	{
		m_lastPartFileStatsSave = dwCurTick;

		SaveAllPartFileStats();
	}
//--- :xrmb ---
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::SaveAllPartFileStats()
{
	EMULE_TRY

	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL;)
	{
		CPartFile	*pPartFile = dynamic_cast<CPartFile*>(m_partFileList.GetNext(pos));

		if (pPartFile)
		{
			pPartFile->SavePartFileStats();
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetFileByIndex() returns the 'iIndex'th (0-based) file in the queue or NULL if 'iIndex' is out of range.
CPartFile *CDownloadQueue::GetFileByIndex(int iIndex)
{
	POSITION	pos = m_partFileList.FindIndex(iIndex);

	return (pos != NULL) ? m_partFileList.GetAt(pos) : NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetFileByID() returns the part file from the download queue with file hash 'strFileHash'
//	or NULL if there is none.
CPartFile *CDownloadQueue::GetFileByID(const uchar* strFileHash)
{
	CPartFile	   *pPartFile = NULL;

	EMULE_TRY

	CPartFile	   *pTempPartFile;

	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL;)
	{
		pTempPartFile = dynamic_cast<CPartFile*>(m_partFileList.GetNext(pos));

		if (pTempPartFile && !md4cmp(strFileHash, pTempPartFile->GetFileHash()))
		{
			pPartFile = pTempPartFile;
			break;
		}
	}

	EMULE_CATCH

	return pPartFile;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CDownloadQueue::IsInDLQueue(CKnownFile *pFileToTest)
{
	bool	bIsInDLQueue = false;

	EMULE_TRY

	POSITION	pos = m_partFileList.Find(pFileToTest);

	if (pos != NULL)
		bIsInDLQueue = true;

	EMULE_CATCH

	return bIsInDLQueue;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CUpDownClient* CDownloadQueue::CheckAndAddSource(CPartFile *pSenderFile, uint32 dwUserIP, uint16 nUserPort, uint32 dwServerIP, uint16 nServerPort, bool bSaveSources)
{
	EMULE_TRY

//	Prevent adding loopback sources (i.e. 127.x.x.x)
	if ((dwUserIP & 0xFF) == 0x7F)
		return NULL;

//	Check if source is our own client (now checking also TCP port)
	if(dwUserIP == g_eMuleApp.m_pGlobPrefs->GetLancastIP() && nUserPort == g_eMuleApp.m_pGlobPrefs->GetPort())
	{
		return NULL;
	}

//	If this part file is stopped or complete/completing or erroneous, no sources needed anymore
	EnumPartFileStatuses	eFileStatus = pSenderFile->GetStatus();

	if (eFileStatus == PS_STOPPED || eFileStatus == PS_COMPLETE
		|| eFileStatus == PS_COMPLETING || eFileStatus == PS_ERROR)
	{
		return NULL;
	}

//	Check if client has to be filtered, to prevent adding it to DL-queue by source exchange
	if (g_eMuleApp.m_pIPFilter->IsFiltered(dwUserIP))
	{
		InterlockedIncrement(&g_eMuleApp.m_lSXFiltered);
		InterlockedIncrement(&g_eMuleApp.m_lTotalFiltered);
		if (!g_eMuleApp.m_pGlobPrefs->IsCMNotLog())
			AddDebugLogLine( RGB_LOG_DIMMED + _T("Filtered (reject as new source): %i.%i.%i.%i %s"),
							 (byte)dwUserIP, (byte)(dwUserIP>>8), (byte)(dwUserIP>>16),
							 (byte)(dwUserIP>>24), g_eMuleApp.m_pIPFilter->GetLastHit() );
		return NULL;
	}

// Now we need to check if client is already in the DL-queue which based on following ideas:
// 1. this function will be always called after object creation (only IP & port are known)
// 2. if client exist as source, he must be presented in ClientList
	CUpDownClient	*pTempSource = g_eMuleApp.m_pClientList->FindClientByPortAndID(dwUserIP,nUserPort);

//	Client is already known
	if (pTempSource)
	{
	//	Check UL only case (client exist only in UL-queue)
		if (!pTempSource->m_pReqPartFile)
		{

		// 	Add if OtherRequestList is empty (source was removed)
			if (pTempSource->m_otherRequestsList.IsEmpty())
			{
				pTempSource->m_pReqPartFile = pSenderFile;
				pSenderFile->m_srcLists[pTempSource->m_iSourcesSlot].AddTail(pTempSource);
				pTempSource->ResetLastAskedTime();
				pSenderFile->RemovePastComment(pTempSource);
				g_eMuleApp.m_pDownloadList->AddSource(pSenderFile,pTempSource,false);
			}
		}
		else if (pTempSource->m_pReqPartFile != pSenderFile)
		{
			if (pTempSource->AddRequestForAnotherFile(pSenderFile))
				g_eMuleApp.m_pDownloadList->AddSource(pSenderFile,pTempSource,true);
		}

		return NULL;
	}
//	Client doesn't exist
	else
	{
		CUpDownClient* pNewSource = new CUpDownClient(nUserPort,dwUserIP,dwServerIP,nServerPort,pSenderFile);

		if (g_eMuleApp.m_pClientList->AddClient(pNewSource,true))
		{
			pNewSource->m_pReqPartFile = pSenderFile;
			pSenderFile->m_srcLists[pNewSource->m_iSourcesSlot].AddTail(pNewSource);
			pSenderFile->RemovePastComment(pNewSource);
			//pNewSource->StartDLQueueWaitTimer();
			g_eMuleApp.m_pDownloadList->AddSource(pSenderFile,pNewSource,false);
			return pNewSource;
		}
		else
			safe_delete(pNewSource);
	}

	EMULE_CATCH

	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CDownloadQueue::CheckAndAddKnownSource(CPartFile* pPartFile,CUpDownClient* pKnownSource)
{
	EMULE_TRY

	EnumPartFileStatuses	eFileStatus = pPartFile->GetStatus();

//	Stop adding sources to stopped, completing, completed and erroneous files
	if (eFileStatus == PS_STOPPED || eFileStatus == PS_COMPLETE
		|| eFileStatus == PS_COMPLETING || eFileStatus == PS_ERROR)
	{
		return FALSE;
	}

	if(!pKnownSource->HasLowID() && (pKnownSource->GetUserID() & 0xFF) == 0x7F)
		return FALSE;

//	If we receive ourselves (now checking IP and UDP port) as a source we drop it
	if ( md4cmp(g_eMuleApp.m_pGlobPrefs->GetUserHash(), pKnownSource->GetUserHash()) == 0 &&
		 pKnownSource->GetIP() == g_eMuleApp.m_pGlobPrefs->GetLancastIP() &&
		 pKnownSource->GetUDPPort() == g_eMuleApp.m_pGlobPrefs->GetUDPPort() )
		return FALSE;

//	Check if file is in download list
	if (IsInDLQueue(pPartFile))
	{
	//	If requested file is already specified
		if (pKnownSource->m_pReqPartFile)
		{
		//	Check if it's the same file
			if (pPartFile == pKnownSource->m_pReqPartFile)
			{
			//	If file was found, to speed up checks we will use variable pKnownSourceslot
				if (pPartFile->m_srcLists[pKnownSource->m_iSourcesSlot].Find(pKnownSource))
					return true;
			}
			else
			{
				if (pKnownSource->AddRequestForAnotherFile(pPartFile))
					g_eMuleApp.m_pDownloadList->AddSource(pPartFile,pKnownSource,false);
				return true;
			}
		}
	}

	pKnownSource->m_pReqPartFile = pPartFile;
	pPartFile->m_srcLists[pKnownSource->m_iSourcesSlot].AddTail(pKnownSource);
	pPartFile->RemovePastComment(pKnownSource);
	//pKnownSource->StartDLQueueWaitTimer();
	g_eMuleApp.m_pDownloadList->AddSource(pPartFile,pKnownSource,false);

	return TRUE;

	EMULE_CATCH

	return FALSE;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// This function removes the source from all files
bool CDownloadQueue::RemoveSource(CUpDownClient* toremove, bool updatewindow)
{
	EMULE_TRY

	bool removed = false;
	CPartFile* cur_file;
//	Set client state to DS_NONE in order to prevent any action from him
	toremove->SetDownloadState(DS_NONE);

//	Check actual requested file (we have this info in "m_pReqPartFile")
	cur_file = toremove->m_pReqPartFile;

	if (cur_file != NULL && !cur_file->m_srcLists[toremove->m_iSourcesSlot].IsEmpty())
	{
		POSITION remove_pos = cur_file->m_srcLists[toremove->m_iSourcesSlot].Find(toremove);
		if (remove_pos != NULL)
		{
			cur_file->AddPastComment(toremove);
			cur_file->m_srcLists[toremove->m_iSourcesSlot].RemoveAt(remove_pos);
			removed = true;
		}
		cur_file->NewSrcPartsInfo();
	}

//	Now remove client in A4AF files. Lets use "OtherRequests_list" for this matter
	for (POSITION pos = toremove->m_otherRequestsList.GetHeadPosition(); pos != NULL; )
	{
		cur_file = toremove->m_otherRequestsList.GetNext(pos);
		if (IsInDLQueue(cur_file))
		{
			POSITION pos_A4AF = cur_file->m_A4AFsrcList.Find(toremove);
			if (pos_A4AF)
			{
				cur_file->m_A4AFsrcList.RemoveAt(pos_A4AF);
			}
		}
	}
//	Clear client from all graphical lists
	if (updatewindow)
		g_eMuleApp.m_pDownloadList->RemoveSource(toremove, NULL);

	if (toremove->GetFileComment().GetLength() > 0 || toremove->GetFileRating() != PF_RATING_NONE)
		toremove->m_pReqPartFile->UpdateFileRatingCommentAvail();

	toremove->m_pReqPartFile = NULL;
	toremove->m_otherNoNeededMap.RemoveAll();
	toremove->m_otherRequestsList.RemoveAll();

	return removed;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::RemoveFile(CPartFile* toremove)
{
	EMULE_TRY

	POSITION remove_pos = m_partFileList.Find(toremove);
	if (remove_pos != NULL)
		m_partFileList.RemoveAt(remove_pos);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::ResumeFiles()
{
	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL; )
	{
		CPartFile* pPartFile = dynamic_cast<CPartFile*>(m_partFileList.GetNext(pos));

		if (pPartFile)
		{
			EnumPartFileStatuses	eFileStatus = pPartFile->GetStatus();

			if ((eFileStatus == PS_READY) || (eFileStatus == PS_EMPTY))
			{
				pPartFile->ResumeFile();
			}
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::DeleteAll()
{
	EMULE_TRY

	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL; )
	{
		CPartFile* pPartFile = dynamic_cast<CPartFile*>(m_partFileList.GetNext(pos));
		if (pPartFile)
		{
			for (int sl = 0;sl < SOURCESSLOTS;sl++)
			{
				if (!pPartFile->m_srcLists[sl].IsEmpty())
					pPartFile->m_srcLists[sl].RemoveAll();
			}
		//	Barry - Should also remove all requested blocks
		//	Don't worry about deleting the blocks, that gets handled
		//	when CUpDownClient is deleted in CClientList::DeleteAll()
			pPartFile->RemoveAllRequestedBlocks();
		}
	}
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Max. file IDs per UDP packet
// ----------------------------
// 576 - 30 bytes of header (28 for UDP, 2 for "E3 9A" edonkey proto) = 546 bytes
// 546 / 16 = 34
bool CDownloadQueue::SendNextUDPPacket()
{
#ifdef OLD_SOCKETS_ENABLED
	EMULE_TRY

	if (!m_bIsInitialized)
		return false;
	if (m_partFileList.IsEmpty() || !g_eMuleApp.m_pServerConnect->IsConnected())
		return false;

//	Get server for request
	pCurUDPServer = g_eMuleApp.m_pServerList->GetNextServer(pCurUDPServer);
//	Check local server & try switch to another one
	if (pCurUDPServer == g_eMuleApp.m_pServerList->GetServerByAddress(g_eMuleApp.m_pServerConnect->GetCurrentServer()->GetAddress(), g_eMuleApp.m_pServerConnect->GetCurrentServer()->GetPort()))
	{
		pCurUDPServer = g_eMuleApp.m_pServerList->GetNextServer(pCurUDPServer);
	}
//	If we got NULL pointer back, that means
//	1. list is empty
//	2. we are in the end of the list
//	in both case we will stop any UDP request
	if (pCurUDPServer == NULL)
	{
		StopUDPRequests();
		return false;
	}

//	Check support of extended request
	int				iMaxFilesPerPacket = 1;

	if (pCurUDPServer->GetUDPFlags() & SRV_UDPFLG_EXT_GETSOURCES)
		iMaxFilesPerPacket = 34;

	CSafeMemFile	dataGlobGetSources(16);
	POSITION		posLastSearchedFile = NULL;

//	Get position in the file list
	if (m_pLastUDPSearchedFile != NULL)
		posLastSearchedFile = m_partFileList.Find(m_pLastUDPSearchedFile);

//	Check if file was found. This can happens in 2 cases
//	1. initial case: m_pLastUDPSearchedFile is not yet definded
//	2. if "m_pLastUDPSearchedFile" was deleted & m_partFileList.Find() return a 0
	if (posLastSearchedFile == NULL)
		posLastSearchedFile = m_partFileList.GetHeadPosition();

//	At this point we have a start position to search in "posLastSearchedFile"
//	so we can create a packet
	int	iRequestedFiles = 0;

	for (int i = 0; i < m_partFileList.GetCount(); i++)
	{
		CPartFile	   *pNextPartFile = m_partFileList.GetNext(posLastSearchedFile);

		if (pNextPartFile)
		{
			EnumPartFileStatuses	eFileStatus = pNextPartFile->GetStatus();

			if (((eFileStatus == PS_READY) || (eFileStatus == PS_EMPTY))
				&& g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFileUDP() > pNextPartFile->GetSourceCount())
			{
				dataGlobGetSources.Write(pNextPartFile->GetFileHash(), 16);
				m_pLastUDPSearchedFile = pNextPartFile;
				iRequestedFiles++;
			}

		//	If packet is filled break the loop
			if (iRequestedFiles >= iMaxFilesPerPacket)
				break;

		//	Check for the end of the loop & start from the head if it's needed
			if (posLastSearchedFile == NULL)
				posLastSearchedFile = m_partFileList.GetHeadPosition();
		}
	}

	Packet			packet(&dataGlobGetSources);

	packet.m_eOpcode = OP_GLOBGETSOURCES;
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadServer(packet.m_dwSize);
	g_eMuleApp.m_pServerConnect->SendUDPPacket(&packet, pCurUDPServer, false);
	return true;

	EMULE_CATCH
#endif //OLD_SOCKETS_ENABLED
	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::StopUDPRequests()
{
	EMULE_TRY

	pCurUDPServer = NULL;
	m_dwLastUDPSearchTime = ::GetTickCount();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::SortByPriority()
{
	EMULE_TRY

	POSITION		pos1, pos2;
	CPartFile	   *pPartFile;
	uint32			i = 0;

	for (pos1 = m_partFileList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		pPartFile = m_partFileList.GetNext(pos1);

		if (pPartFile)
		{
			if (pPartFile->GetPriority() == PR_HIGH)
			{
				m_partFileList.AddHead(pPartFile);
				m_partFileList.RemoveAt(pos2);
			}
			else if (pPartFile->GetPriority() == PR_LOW)
			{
				m_partFileList.AddTail(pPartFile);
				m_partFileList.RemoveAt(pos2);
			}

			if (++i == m_partFileList.GetCount())
				break;
		}
	}
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CUpDownClient* CDownloadQueue::GetDownloadClientByIP(uint32 dwIP)
{
	EMULE_TRY

	POSITION			pos1, pos2;
	CUpDownClient	   *pSource;
	CPartFile		   *pPartFile;

	for (pos1 = m_partFileList.GetHeadPosition(); pos1 != NULL;)
	{
		pPartFile = m_partFileList.GetNext(pos1);

		if (pPartFile)
		{
			for (int sl = 0; sl < SOURCESSLOTS; sl++)
			{
				for (pos2 = pPartFile->m_srcLists[sl].GetHeadPosition(); pos2 != NULL;)
				{
					pSource = pPartFile->m_srcLists[sl].GetNext(pos2);

					if (pSource && dwIP == pSource->GetIP())
					{
						return pSource;
					}
				}
			}
		}
	}
	EMULE_CATCH
	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CUpDownClient* CDownloadQueue::GetDownloadClientByIP_UDP(uint32 dwIP, uint16 nUDPPort)
{
	EMULE_TRY

	POSITION			pos1, pos2;
	CUpDownClient	   *pSource;
	CPartFile		   *pPartFile;

	for (pos1 = m_partFileList.GetHeadPosition(); pos1 != NULL;)
	{
		pPartFile = m_partFileList.GetNext(pos1);

		if (pPartFile)
		{
			for (int sl = 0; sl < SOURCESSLOTS; sl++)
			{
				for (pos2 = pPartFile->m_srcLists[sl].GetHeadPosition(); pos2 != NULL;)
				{
					pSource = pPartFile->m_srcLists[sl].GetNext(pos2);

					if (pSource && dwIP == pSource->GetIP() && nUDPPort == pSource->GetUDPPort())
					{
						return pSource;
					}
				}
			}
		}
	}
	EMULE_CATCH
	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// WARNING: by adding a new type don't forget to change constants in opcodes
void CDownloadQueue::GetDownloadStats(uint16 sources[], uint64 data[])
{
	EMULE_TRY

	memzero(sources, sizeof(sources[0]) * STATS_DL_SOURCES_TYPE);
	memzero(data, sizeof(data[0]) * STATS_DL_DATA_INFO);

	CPartFile	   *pPartFile;

	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL; )
	{
		pPartFile = m_partFileList.GetNext(pos);
		if (pPartFile)
		{
			sources[0] += pPartFile->GetSourceCount();
			sources[1] += pPartFile->GetTransferringSrcCount();
		//	Extra stats
			sources[2] += pPartFile->GetOnQueueSrcCount();
			sources[3] += pPartFile->GetQueueFullSrcCount();
			sources[4] += pPartFile->GetNoNeededPartsSrcCount();
			sources[5] += pPartFile->GetConnectedSrcCount();
			sources[6] += pPartFile->GetRecHashsetSrcCount();
			sources[7] += pPartFile->GetConnectingSrcCount();
			sources[8] += pPartFile->GetConnectingViaServerSrcCount();
			sources[9] += pPartFile->GetTooManyConnsSrcCount();
			sources[10] += pPartFile->GetLow2LowSrcCount();
			sources[11] += pPartFile->GetUnknownSrcCount();
			sources[12] += pPartFile->GetErrorSrcCount();
			sources[13] += pPartFile->GetBannedSrcCount();
			sources[14] += pPartFile->TotalPacketsSavedDueToICH();
			sources[15] += pPartFile->GetSrcA4AFCount();
			sources[16] += pPartFile->GetTimeoutedSrcCount();
			sources[17] += pPartFile->GetHighQRSrcCount();
			sources[18] += pPartFile->GetNoUploadSrcCount();
		//	Some data
			data[0] += pPartFile->GetSessionLostDueToCorruption();
			data[1] += pPartFile->GetSessionGainDueToCompression();

			uint32 dwSizeToTransfer = 0;
			uint32 dwNeededSpace = (pPartFile->GetFileSize() - pPartFile->GetRealFileSize());
			EnumPartFileStatuses	eFileStatus = pPartFile->GetStatus();

			pPartFile->GetSizeToTransfer(dwSizeToTransfer);
			if ((eFileStatus ==  PS_READY) || (eFileStatus == PS_EMPTY))
			{
				data[2] += pPartFile->GetFileSize();
				data[3] += dwSizeToTransfer;
				data[4] += dwNeededSpace;
			}
			data[5] += pPartFile->GetFileSize();
			data[6] += dwNeededSpace;
			data[7] += pPartFile->GetLostDueToCorruption();
			data[8] += pPartFile->GetGainDueToCompression();
			data[9] += pPartFile->GetRealFileSize();
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::SetAutoSourcesPerFile()
{
	uint32 cur_tmc = 0;
	uint32 cur_stotal = 0;
	uint32 old_mspf = g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile();
	uint32 new_mspf = old_mspf;
	uint32 min_spf = g_eMuleApp.m_pGlobPrefs->GetMinAutoSourcesPerFile();
	uint32 max_spf = g_eMuleApp.m_pGlobPrefs->GetMaxAutoSourcesPerFile();
	uint32 max_stotal = g_eMuleApp.m_pGlobPrefs->GetMaxAutoSourcesTotal();
	uint32 max_sesrc = g_eMuleApp.m_pGlobPrefs->GetMaxAutoExchangeSources();
//	Count too many connections for all files
	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != 0; )
	{
		CPartFile * cur_file = m_partFileList.GetNext(pos);
		cur_tmc += cur_file->GetTooManyConnsSrcCount();
		cur_stotal += cur_file->GetSourceCount();
	}

//	Check lower and upper limit
	uint32 current_dls = g_eMuleApp.m_pDownloadQueue->GetActiveFileCount();
	if (current_dls > 0)
	{
		uint32 useful_min = max_stotal / current_dls;
		if (min_spf > useful_min)
		{
			min_spf = useful_min;
			g_eMuleApp.m_pGlobPrefs->SetMinAutoSourcesPerFile(min_spf);
		}
	}
	if (max_spf > max_stotal)
	{
		max_spf = max_stotal;
		g_eMuleApp.m_pGlobPrefs->SetMaxAutoSourcesPerFile(max_spf);
	}

//	Adjust SPF according to situation
	if (cur_stotal > max_stotal)
	{
	//	Too many total sources, lower
		if (cur_stotal - max_stotal > 200)
		{
			new_mspf = old_mspf - 10;
		}
		else
		{
			new_mspf = old_mspf - 2;
		}
	}
	else
	{
	//	Too many connections, lower sources per file
		if ((cur_tmc > 10) && (old_mspf > min_spf))
		{
			new_mspf = old_mspf - (cur_tmc / 15);
		}

	//	Not too many connections, raise sources per file depending on situation
		else if ((cur_tmc < 11) && (old_mspf < 200))
		{
			new_mspf = old_mspf + 10;
		}
		else if ((cur_tmc < 21) && (old_mspf < 200))
		{
			new_mspf = old_mspf + 2;
		}
		else if ((cur_tmc < 6) && (old_mspf < 400))
		{
			new_mspf = old_mspf + 5;
		}
		else if ((cur_tmc < 11) && (old_mspf < 400))
		{
			new_mspf = old_mspf + 2;
		}
		else if ((cur_tmc < 6) && (old_mspf < 600))
		{
			new_mspf = old_mspf + 5;
		}
		else if ((cur_tmc < 2) && (old_mspf < 800))
		{
			new_mspf = old_mspf + 5;
		}
		else if ((cur_tmc == 0) && (old_mspf < max_spf))
		{
			new_mspf = old_mspf + 2;
		}
	}

//	Stay between min_spf and max_spf
	if (new_mspf < min_spf)
	{
		new_mspf = min_spf;
	}
	if (new_mspf > max_spf)
	{
		new_mspf = max_spf;
	}

//	Change values
	if (old_mspf != new_mspf)
	{
		g_eMuleApp.m_pGlobPrefs->SetMaxSourcePerFile(new_mspf);

		if (g_eMuleApp.m_pGlobPrefs->IsAutoSourcesLogEnabled())
		{
			AddDebugLogLine( RGB_LOG_NOTICE + _T("AutoSourcesPerFile: TMC=%u / Total=%u -> Changed from %u to %u (%+d)"),
							 cur_tmc, cur_stotal, old_mspf, new_mspf, new_mspf-old_mspf );
		}

	//	Change source excange limit according to sources per file
		uint32 new_sesrc = g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFileSoft();
		if (new_sesrc < max_sesrc)
		{
			g_eMuleApp.m_pGlobPrefs->SetXSUpTo(new_sesrc);
		}
		else
		{
			g_eMuleApp.m_pGlobPrefs->SetXSUpTo(max_sesrc);
		}
	}
	else if (g_eMuleApp.m_pGlobPrefs->IsAutoSourcesLogEnabled())
	{
		AddDebugLogLine( RGB_LOG_NOTICE + _T("AutoSourcesPerFile: TMC=%u / Total=%u -> Not changed from %u"),
						 cur_tmc, cur_stotal, old_mspf );
	}
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CDownloadQueue::GetTransferringFiles() const
{
	EMULE_TRY

	int				iCount = 0;
	CPartFile	   *pPartFile;

	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL;)
	{
		pPartFile = m_partFileList.GetNext(pos);

		if (pPartFile)
		{
			if (pPartFile->GetTransferringSrcCount() > 0)
				iCount++;
		}
	}

	return iCount;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CDownloadQueue::GetPausedFileCount() const
{
	EMULE_TRY

	int				iCount = 0;
	CPartFile	   *pPartFile;

	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL;)
	{
		pPartFile = m_partFileList.GetNext(pos);

		if (pPartFile)
		{
			if (pPartFile->GetStatus() == PS_PAUSED)
				iCount++;
		}
	}

	return iCount;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CDownloadQueue::GetStoppedFileCount() const
{
	EMULE_TRY

	int				iCount = 0;
	CPartFile	   *pPartFile;

	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL;)
	{
		pPartFile = m_partFileList.GetNext(pos);

		if (pPartFile != NULL && pPartFile->GetStatus() == PS_STOPPED)
		{
			iCount++;
		}
	}

	return iCount;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CDownloadQueue::GetActiveFileCount() const
{
	EMULE_TRY

	int				iCount = 0;
	CPartFile	   *pPartFile;

	for (POSITION pos = m_partFileList.GetHeadPosition(); pos != NULL;)
	{
		pPartFile = m_partFileList.GetNext(pos);

		if (pPartFile)
		{
			EnumPartFileStatuses	eFileStatus = pPartFile->GetStatus();

			if ((eFileStatus == PS_READY) || (eFileStatus == PS_EMPTY))
			{
				iCount++;
			}
		}
	}

	return iCount;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::AddClientHostnameToResolve(CTypedPtrList<CPtrList, CClientSource*>* pLink)
{
	if (pLink)
		hostnameResolveQueue.AddTail(pLink);

	if (hostnameResolveQueue.IsEmpty())
		return;

	if (m_bIsResolving)
	{
		return;
	}

	ResolveNextSourceHostname();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::ResolveNextSourceHostname()
{
	m_bIsResolving = true;
	CClientSource* source;
	for (;;)
	{
	 //	No more hostnames to resolve
		if (hostnameResolveQueue.IsEmpty())
		{
			if (m_wndHSCallback)
			{
			//	Free unneccessary window
				delete m_wndHSCallback;
				m_wndHSCallback = NULL;
			}
			m_bIsResolving = false;
			return ;
		}

		source = hostnameResolveQueue.GetHead();

		if (source->sourceType != ED2KLINK_SOURCE_HOSTNAME)
		{
			delete source;
			hostnameResolveQueue.RemoveHead();
		}
		else
			break;
	}

	memzero(hostentBuffer, sizeof(hostentBuffer));

	if (!m_wndHSCallback)
	{
		m_wndHSCallback = new CHostnameSourceWnd();
		VERIFY(m_wndHSCallback->CreateEx(0, AfxRegisterWndClass(0), _T("Emule DNS Source Socket Wnd"), WS_OVERLAPPED, 0, 0, 0, 0, NULL, NULL));
		m_wndHSCallback->m_pOwner = this;
	}

	AddDebugLogLine(_T("%s: Resolving source hostname: %s"), __FUNCTION__, source->sourceHostname);
	HANDLE RetValue;
	USES_CONVERSION;
	RetValue = WSAAsyncGetHostByName( m_wndHSCallback->m_hWnd, TM_SOURCEHOSTNAMERESOLVED,
	                                  CT2CA(source->sourceHostname), hostentBuffer, MAXGETHOSTSTRUCT );

	if (RetValue == 0)
	{
		if (sourceHostnameResolveRetry > 3)
		{
			CClientSource * hostsource;
			hostsource = hostnameResolveQueue.RemoveHead();
			AddDebugLogLine( RGB_LOG_ERROR + _T("%s: Error calling WSAAsyncGetHostByName to resolve source '%s'"),
							 __FUNCTION__, hostsource->sourceHostname );
			delete hostsource;
			if (m_wndHSCallback)
			{
				delete m_wndHSCallback;
				m_wndHSCallback = NULL;
			}
			memzero(hostentBuffer, sizeof(hostentBuffer));
			m_bIsResolving = false;
		}
		else
		{
			sourceHostnameResolveRetry += 1;
			memzero(hostentBuffer, sizeof(hostentBuffer));
			AddDebugLogLine(_T("%s: Source resolve retry number %i"), __FUNCTION__, sourceHostnameResolveRetry);
			ResolveNextSourceHostname();
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::SourceHostnameResolved(WPARAM wp, LPARAM lp)
{

	ASSERT (hostnameResolveQueue.GetCount() > 0);

	sourceHostnameResolveRetry = 0;

	CClientSource* hostsource;
	hostsource = hostnameResolveQueue.RemoveHead();

	if (WSAGETASYNCERROR(lp) != 0)
	{
		DEBUG_ONLY(AddDebugLogLine(RGB_LOG_ERROR + _T("%s: Error resolving source '%s'"), __FUNCTION__, hostsource->sourceHostname));
		delete hostsource;
		delete m_wndHSCallback;
		m_wndHSCallback = NULL;
		memzero(hostentBuffer, sizeof(hostentBuffer));
		ResolveNextSourceHostname();
		return ;
	}

	int iBufLen = WSAGETASYNCBUFLEN(lp);
	LPHOSTENT lphost = (LPHOSTENT)malloc(iBufLen);
	memcpy2(lphost, hostentBuffer, iBufLen);
	hostsource->sourceID = ((LPIN_ADDR)lphost->h_addr_list[0])->s_addr;
	in_addr resolvedaddr = *(LPIN_ADDR)lphost->h_addr_list[0];
	free(lphost);

	AddDebugLogLine( _T("%s: Resolved source hostname: %s - %i:%i:%i:%i"),
					 __FUNCTION__,
					 hostsource->sourceHostname,
					 resolvedaddr.S_un.S_un_b.s_b1,
					 resolvedaddr.S_un.S_un_b.s_b2,
					 resolvedaddr.S_un.S_un_b.s_b3,
					 resolvedaddr.S_un.S_un_b.s_b4 );
	CPartFile* partfile = this->GetFileByID(hostsource->filehashkey);
	if (partfile)
		partfile->AddClientSource(hostsource);
	else
	{
	//	TODO: notify user the partfile exists no more?
		DEBUG_ONLY(AddDebugLogLine( RGB_LOG_WARNING + _T("%s: impossible to add source because the partfile doesn't exist anymore")
									__FUNCTION__ ));
	}
	delete hostsource;
	memzero(hostentBuffer, sizeof(hostentBuffer));
	ResolveNextSourceHostname();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CDownloadQueue::GetUDPSearchStatus()
{
	EMULE_TRY

	if (!m_bIsInitialized)
		return _T("");
#ifdef OLD_SOCKETS_ENABLED

	if (m_partFileList.IsEmpty() || !g_eMuleApp.m_pServerConnect->IsConnected())
		return _T("");
#endif //OLD_SOCKETS_ENABLED

	CString strOut;

	if (pCurUDPServer)
	{
		strOut = GetResString(IDS_UDPSEARCH_PROGRESS);
	}
	else if (m_dwLastUDPSearchTime)
	{
		uint32 dwNextUDPSearchTimeInSec = (m_dwLastUDPSearchTime + UDPSERVERREASKTIME - ::GetTickCount())/1000;
		strOut.Format( GetResString(IDS_UDPSEARCH_NEXT), CastSecondsToHM(dwNextUDPSearchTimeInSec));
	}
	else
		strOut = GetResString(IDS_UDPSEARCH_NOTSTART);

	return strOut;

	EMULE_CATCH
	return _T("");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::SetA4AFAutoFile(CPartFile *file)
{
	EMULE_TRY

	m_A4AF_auto_file = file;

	if (file)
		g_eMuleApp.m_pGlobPrefs->SetA4AFHash(file->GetFileHash());
	else
	{
	//	Set null hash
		uchar null_hash[16];

		md4clr(null_hash);
		g_eMuleApp.m_pGlobPrefs->SetA4AFHash(null_hash);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::ResetCatParts(int iCatIndex)
{
//	We need a copy of the list because of the original list update done in SetCatID
	CTypedPtrList<CPtrList, CPartFile*> locallist;

	POSITION		pos;
	CPartFile	   *pPartFile;

	for (pos = m_partFileList.GetHeadPosition(); pos != NULL; )
	{
		pPartFile = m_partFileList.GetNext(pos);
		if (pPartFile)
		{
			locallist.AddTail(pPartFile);
		}
	}

	for (pos = locallist.GetHeadPosition(); pos != NULL; )
	{
		pPartFile = locallist.GetNext(pos);

		if (CCat::GetCatIndexByID(pPartFile->GetCatID()) == iCatIndex)
			pPartFile->SetCatID(CAT_NONE);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::SetCatPrio(int iCatIndex, byte newprio)
{
//	We need a copy of the list because of the original list update done in SetCatID
	CTypedPtrList<CPtrList, CPartFile*>	locallist;

	POSITION		pos;
	CPartFile	   *pPartFile;

	for (pos = m_partFileList.GetHeadPosition(); pos != NULL; )
	{
		pPartFile = m_partFileList.GetNext(pos);
		if (pPartFile)
		{
			locallist.AddTail(pPartFile);
		}
	}

	for (pos = locallist.GetHeadPosition(); pos != NULL; )
	{
		pPartFile = locallist.GetNext(pos);

		if (iCatIndex == 0 || CCat::GetCatIndexByID(pPartFile->GetCatID()) == iCatIndex)
		{
			if (newprio == PR_AUTO)
			{
				pPartFile->SetAutoPriority(true);
				pPartFile->UpdateDownloadAutoPriority();
			}
			else
			{
				pPartFile->SetAutoPriority(false);
				pPartFile->SetPriority(newprio);
			}
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::SetCatStatus(int iCatIndex, uint16 newstatus)
{
	bool reset = false;
	POSITION pos = m_partFileList.GetHeadPosition();

	while (pos != NULL)
	{
		CPartFile * cur_file = m_partFileList.GetNext(pos);

		if (!cur_file)
			continue;

		if ( iCatIndex == -1 || (iCatIndex == -2 && cur_file->GetCatID() == 0)
		     || (iCatIndex == 0 && CCat::FileBelongsToGivenCat(cur_file, CAT_ALL))
		     || (iCatIndex >= CCat::GetNumPredefinedCats() && CCat::GetCatIDByIndex(iCatIndex) == cur_file->GetCatID()) )
		{
			switch (newstatus)
			{
				case MP_CANCEL:
					cur_file->DeleteFile();
					reset = true;
					break;
				case MP_PAUSE:
					cur_file->PauseFile();
					break;
				case MP_STOP:
					cur_file->StopFile();
					break;
				case MP_RESUME:
					if(cur_file->GetStatus() == PS_PAUSED || cur_file->GetStatus() == PS_STOPPED)
						cur_file->ResumeFile();
					break;
			}
		}
		if (reset)
		{
			reset = false;
			pos = m_partFileList.GetHeadPosition();
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CDownloadQueue::SetAutoCat(CPartFile *newfile)
{
	if (CCat::GetNumCats() > 1)
	{
		for (int ix = CCat::GetNumPredefinedCats(); ix < CCat::GetNumCats(); ix++)
		{
			int			curPos = 0;
			CString		catExt = CCat::GetCatByIndex(ix)->GetAutoCatExt();

		//	No need to compare again an empty AutoCat array
			if (catExt == "")
				continue;

			catExt.MakeLower();

			CString		fullname = newfile->GetFileName();

			fullname.MakeLower();

			CString		cmpExt = catExt.Tokenize("|", curPos);

			while (cmpExt != "")
			{
			//	Allow wildcards in autocat string
				if (cmpExt.FindOneOf("*?") != -1)
				{
				//	Use wildcards
					char * file = fullname.GetBuffer();
					char* spec = cmpExt.GetBuffer();
					if (PathMatchSpec(file, spec))
					{
						if(newfile->GetCatID() == CAT_NONE)
							newfile->SetCatID(CCat::GetCatIDByIndex(ix));
						return ;
					}
				}
				else
				{
					if (fullname.Find(cmpExt) != -1)
					{
						if(newfile->GetCatID() == CAT_NONE)
							newfile->SetCatID(CCat::GetCatIDByIndex(ix));
						return ;
					}
				}
				cmpExt = catExt.Tokenize("|", curPos);
			}
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
