/////////////////////////////////////////////////////////
// EmMsgs.h - definition of all eMule messages (opcodes)
//
// NOTE Don't put <#pragma once> directive here. 
// This file should be included multiple times.



/////////////////////////////////////////////////////////
// Client OpCodes 
/////////////////////////////////////////////////////////

// Both HELLO and HELLOANSWER have exactly the same parameters.
#ifndef OPCODE_ENDONKEY_HELLO_PARAMS
#	define OPCODE_ENDONKEY_HELLO_PARAMS \
		PARAM_BYTE_JUNK(16) \
		PARAM_HASH(UserHash) \
		PARAM_ADDRPORT(ClientAddr) \
		PARAM_TAGS_BEGIN \
			PARAM_TAG_STR(UserName, CT_NAME) \
			PARAM_TAG_DWORD(ClientSoftVersion, CT_VERSION) \
			PARAM_TAG_DWORD(UserPort, CT_PORT) \
			PARAM_TAG_STR(ModVersion, CT_MOD_VERSION) \
		PARAM_TAGS_END \
		PARAM_ADDRPORT(ServerAddr) \

#endif // OPCODE_ENDONKEY_HELLO_PARAMS

/////////////////////////////////////////////////////////
// Hello OpCode, sent right after connecting to client
BEGIN_OPCODE(0x01, HELLO, EDONKEY, PEER)
	OPCODE_ENDONKEY_HELLO_PARAMS
END_OPCODE

/////////////////////////////////////////////////////////
// Answer to Hello OpCode
BEGIN_OPCODE(0x4C, HELLOANSWER, EDONKEY, PEER)
	OPCODE_ENDONKEY_HELLO_PARAMS
END_OPCODE


// Both EMULEINFO and EMULEINFOANSWER have exactly the same parameters.
#ifndef OPCODE_EMULE_EMULEINFO_PARAMS
#	define OPCODE_EMULE_EMULEINFO_PARAMS \
		PARAM_BYTE(VersionClient) \
		PARAM_BYTE(VersionProtocol) \
		PARAM_TAGS_BEGIN \
			PARAM_TAG_DWORD(Compression, ET_COMPRESSION) \
			PARAM_TAG_DWORD(UdpVersion, ET_UDPVER) \
			PARAM_TAG_DWORD(UdpPort, ET_UDPPORT) \
			PARAM_TAG_DWORD(SourceExchange, ET_SOURCEEXCHANGE) \
			PARAM_TAG_DWORD(Comments, ET_COMMENTS) \
			PARAM_TAG_DWORD(ModPlus, ET_MOD_PLUS) \
			PARAM_TAG_DWORD(ExtendedRequest, ET_EXTENDEDREQUEST) \
			PARAM_TAG_DWORD(ModVersion, ET_MOD_VERSION) \
			PARAM_TAG_DWORD(L2Hac, ET_L2HAC) \
			PARAM_TAG_DWORD(Features, ET_FEATURES) \
 		PARAM_TAGS_END \
		PARAM_ADDRPORT(ServerAddr) \

#endif // OPCODE_EMULE_EMULEINFO_PARAMS


/////////////////////////////////////////////////////////
// EmuleInfo OpCode
BEGIN_OPCODE(0x01, EMULEINFO, EMULE, PEER)
	OPCODE_EMULE_EMULEINFO_PARAMS
END_OPCODE

BEGIN_OPCODE(0x02, EMULEINFOANSWER, EMULE, PEER)
	OPCODE_EMULE_EMULEINFO_PARAMS
END_OPCODE


/////////////////////////////////////////////////////////
// Offer files
BEGIN_OPCODE(0x15, OFFERFILES, EDONKEY, PEER)

// implement complex array later
	PARAM_COMPLEXARRAY_BEGIN(OFFERFILES, Files)
		PARAM_HASH(Hash)
		PARAM_ADDRPORT(ClientAddr)

		PARAM_TAGS_BEGIN
			PARAM_TAG_STR(FileName, FT_FILENAME)
			PARAM_TAG_DWORD(FileSize, FT_FILESIZE)
		PARAM_TAGS_END

	PARAM_COMPLEXARRAY_END(Files)

END_OPCODE


/////////////////////////////////////////////////////////
// Server OpCodes
/////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////
// Login request
BEGIN_OPCODE(0x01, LOGINREQUEST, EDONKEY, SERVER)

	PARAM_HASH(UserHash)
	PARAM_DWORD(ClientID)		// Why not ADDRPORT? Because LowID isn't IP? Isn't that true everywhere?
	PARAM_USHORT(ClientPort)

	PARAM_TAGS_BEGIN
		PARAM_TAG_STR(UserName, CT_NAME)
		PARAM_TAG_DWORD(ClientSoftVersion, CT_VERSION)
		PARAM_TAG_DWORD(UserPort, CT_PORT)
	PARAM_TAGS_END

END_OPCODE

/////////////////////////////////////////////////////////
// Server message
BEGIN_OPCODE(0x38, SERVERMESSAGE, EDONKEY, SERVER)

	PARAM_STRING(Msg)

	#ifdef PARAM_INTERNAL_DEFINITION // Internal variables
	int			m_nMsgCount;
	#endif // PARAM_INTERNAL_DEFINITION

END_OPCODE

/////////////////////////////////////////////////////////
// ID change
BEGIN_OPCODE(0x40, IDCHANGE, EDONKEY, SERVER)

	PARAM_DWORD(NewClientID)
	PARAM_DWORD(SupportedFeatures)

END_OPCODE

/////////////////////////////////////////////////////////
// Search results
BEGIN_OPCODE(0x33, SEARCHRESULT, EDONKEY, SERVER)

	// search results

END_OPCODE

/////////////////////////////////////////////////////////
// Found sources
BEGIN_OPCODE(0x42, FOUNDSOURCES, EDONKEY, SERVER)

	PARAM_HASH(Hash)
	PARAM_ARRAY(Sources, ADDRPORT, BYTE)

END_OPCODE

/////////////////////////////////////////////////////////
// Server status update
BEGIN_OPCODE(0x34, SERVERSTATUS, EDONKEY, SERVER)

	PARAM_DWORD(NumberOfUsers)
	PARAM_DWORD(NumberOfFiles)

END_OPCODE

/////////////////////////////////////////////////////////
// Server info
BEGIN_OPCODE(0x41, SERVERIDENT, EDONKEY, SERVER)

	PARAM_HASH(ServerHash)
	PARAM_ADDRPORT(ServerAddr)

	PARAM_TAGS_BEGIN
		PARAM_TAG_STR(ServerName, ST_SERVERNAME)
		PARAM_TAG_STR(ServerDescription, ST_DESCRIPTION)
	PARAM_TAGS_END

END_OPCODE

/////////////////////////////////////////////////////////
// Server list
BEGIN_OPCODE(0x32, SERVERLIST, EDONKEY, SERVER)

	PARAM_ARRAY(Servers, ADDRPORT, BYTE)

END_OPCODE

/////////////////////////////////////////////////////////
BEGIN_OPCODE(0x1C, CALLBACKREQUEST, EDONKEY, SERVER)

	PARAM_DWORD(UserID)

END_OPCODE

/////////////////////////////////////////////////////////
//	Potential response to CALLBACKREQUEST. Unobserved in
//		the wild ;)
BEGIN_OPCODE(0x36, CALLBACKFAIL, EDONKEY, SERVER)

END_OPCODE

/////////////////////////////////////////////////////////
// Some client request callback through server
BEGIN_OPCODE(0x35, CALLBACKREQUESTED, EDONKEY, SERVER)

	PARAM_ADDRPORT(ClientAddr)

END_OPCODE
